# -*- coding: utf-8 -*-

#    Copyright (C) 2016 Adam Collin, Mathew Topper
#    Copyright (C) 2017-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Wed Apr 06 15:59:04 2016

.. moduleauthor:: Adam Collin <adam.collin@ieee.org>
.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

from datetime import timedelta
from textwrap import wrap
from typing import Optional

import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.dates import (
    MONTHLY,
    DateFormatter,
    RRuleLocator,
    date2num,
    rrulewrapper,
)

from .base import PlotInterface


class InstallationGanttChartPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Installation Gantt Chart"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "project.install_support_structure_dates",
            "project.install_devices_dates",
            "project.install_dynamic_cable_dates",
            "project.install_export_cable_dates",
            "project.install_array_cable_dates",
            "project.install_surface_piercing_substation_dates",
            "project.install_subsea_collection_point_dates",
            "project.install_cable_protection_dates",
            "project.install_driven_piles_dates",
            "project.install_direct_embedment_dates",
            "project.install_gravity_based_dates",
            "project.install_pile_anchor_dates",
            "project.install_drag_embedment_dates",
            "project.install_suction_embedment_dates",
            "project.device_phase_installation_times",
            "project.electrical_phase_installation_times",
            "project.mooring_phase_installation_times",
            "project.installation_plan",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = [
            "project.install_support_structure_dates",
            "project.install_devices_dates",
            "project.install_dynamic_cable_dates",
            "project.install_export_cable_dates",
            "project.install_array_cable_dates",
            "project.install_surface_piercing_substation_dates",
            "project.install_subsea_collection_point_dates",
            "project.install_cable_protection_dates",
            "project.install_driven_piles_dates",
            "project.install_direct_embedment_dates",
            "project.install_gravity_based_dates",
            "project.install_pile_anchor_dates",
            "project.install_drag_embedment_dates",
            "project.install_suction_embedment_dates",
            "project.device_phase_installation_times",
            "project.electrical_phase_installation_times",
            "project.mooring_phase_installation_times",
            "project.installation_plan",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "install_support_structure_dates": "project.install_support_structure_dates",
            "install_devices_dates": "project.install_devices_dates",
            "install_dynamic_cable_dates": "project.install_dynamic_cable_dates",
            "install_export_cable_dates": "project.install_export_cable_dates",
            "install_array_cable_dates": "project.install_array_cable_dates",
            "install_surface_piercing_substation_dates": "project.install_surface_piercing_substation_dates",
            "install_subsea_collection_point_dates": "project.install_subsea_collection_point_dates",
            "install_cable_protection_dates": "project.install_cable_protection_dates",
            "install_driven_piles_dates": "project.install_driven_piles_dates",
            "install_direct_embedment_dates": "project.install_direct_embedment_dates",
            "install_gravity_based_dates": "project.install_gravity_based_dates",
            "install_pile_anchor_dates": "project.install_pile_anchor_dates",
            "install_drag_embedment_dates": "project.install_drag_embedment_dates",
            "install_suction_embedment_dates": "project.install_suction_embedment_dates",
            "install_device_times": "project.device_phase_installation_times",
            "install_electrical_times": "project.electrical_phase_installation_times",
            "install_mooring_times": "project.mooring_phase_installation_times",
            "plan": "project.installation_plan",
        }

        return id_map

    def connect(self):
        self.fig_handle = installation_gantt_chart(
            self.data.plan,
            self.data.install_support_structure_dates,
            self.data.install_devices_dates,
            self.data.install_dynamic_cable_dates,
            self.data.install_export_cable_dates,
            self.data.install_array_cable_dates,
            self.data.install_surface_piercing_substation_dates,
            self.data.install_subsea_collection_point_dates,
            self.data.install_cable_protection_dates,
            self.data.install_driven_piles_dates,
            self.data.install_direct_embedment_dates,
            self.data.install_gravity_based_dates,
            self.data.install_pile_anchor_dates,
            self.data.install_drag_embedment_dates,
            self.data.install_suction_embedment_dates,
            self.data.install_device_times,
            self.data.install_electrical_times,
            self.data.install_mooring_times,
        )


def installation_gantt_chart(
    plan=None,
    install_support_structure_dates=None,
    install_devices_dates=None,
    install_dynamic_cable_dates=None,
    install_export_cable_dates=None,
    install_array_cable_dates=None,
    install_surface_piercing_substation_dates=None,
    install_subsea_collection_point_dates=None,
    install_cable_protection_dates=None,
    install_driven_piles_dates=None,
    install_direct_embedment_dates=None,
    install_gravity_based_dates=None,
    install_pile_anchor_dates=None,
    install_drag_embedment_dates=None,
    install_suction_embedment_dates=None,
    install_device_times: Optional[pd.DataFrame] = None,
    install_electrical_times: Optional[pd.DataFrame] = None,
    install_mooring_times: Optional[pd.DataFrame] = None,
):
    if plan is None:
        return None

    installation = {}
    # sort data
    if (
        any("support structure" in phase for phase in plan)
        and install_device_times is not None
    ):
        component_time = install_device_times.loc["Support Structure"]

        values = installation_gantt_dates(
            install_support_structure_dates, component_time["Preparation"]
        )

        installation["Installation of support structure"] = values

    if (
        any("devices" in phase for phase in plan)
        and install_device_times is not None
    ):
        component_time = install_device_times.loc["Device"]

        values = installation_gantt_dates(
            install_devices_dates, component_time["Preparation"]
        )

        installation["Installation of devices"] = values

    if (
        any("dynamic" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc["Dynamic Cables"]

        values = installation_gantt_dates(
            install_dynamic_cable_dates, component_time["Preparation"]
        )

        installation["Installation of dynamic cables"] = values

    if (
        any("export" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc["Export Cables"]

        values = installation_gantt_dates(
            install_export_cable_dates, component_time["Preparation"]
        )

        installation["Installation of static export cables"] = values

    if (
        any("array" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc["Inter-Array Cables"]

        values = installation_gantt_dates(
            install_array_cable_dates, component_time["Preparation"]
        )

        installation["Installation of static array cables"] = values

    if (
        any("surface piercing" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc["Collection Points"]

        values = installation_gantt_dates(
            install_surface_piercing_substation_dates,
            component_time["Preparation"],
        )

        installation["Installation of collection point (surface piercing)"] = (
            values
        )

    if (
        any("seabed" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc["Collection Points"]

        values = installation_gantt_dates(
            install_subsea_collection_point_dates, component_time["Preparation"]
        )

        installation["Installation of collection point (seabed)"] = values

    if (
        any("cable protection" in phase for phase in plan)
        and install_electrical_times is not None
    ):
        component_time = install_electrical_times.loc[
            "External Cable Protection"
        ]

        values = installation_gantt_dates(
            install_cable_protection_dates, component_time["Preparation"]
        )

        installation["Installation of external cable protection"] = values

    if (
        any("driven piles" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Driven Piles"]

        values = installation_gantt_dates(
            install_driven_piles_dates, component_time["Preparation"]
        )

        installation["Installation of driven piles anchors/foundations"] = (
            values
        )

    if (
        any("direct-embedment" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Direct-Embedment Anchors"]

        values = installation_gantt_dates(
            install_direct_embedment_dates, component_time["Preparation"]
        )

        installation[
            "Installation of mooring systems with direct-embedment " "anchors"
        ] = values

    if (
        any("gravity based" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Gravity Based Foundations"]

        values = installation_gantt_dates(
            install_gravity_based_dates, component_time["Preparation"]
        )

        installation["Installation of gravity based foundations"] = values

    if (
        any("pile anchor" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Pile Anchors"]

        values = installation_gantt_dates(
            install_pile_anchor_dates, component_time["Preparation"]
        )

        installation["Installation of mooring systems with pile anchors"] = (
            values
        )

    if (
        any("drag-embedment" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Drag-Embedment Anchors"]

        values = installation_gantt_dates(
            install_drag_embedment_dates, component_time["Preparation"]
        )

        installation[
            "Installation of mooring systems with drag-embedment " "anchors"
        ] = values

    if (
        any("suction-embedment" in phase for phase in plan)
        and install_mooring_times is not None
    ):
        component_time = install_mooring_times.loc["Suction-Caisson Anchors"]

        values = installation_gantt_dates(
            install_suction_embedment_dates, component_time["Preparation"]
        )

        installation[
            "Installation of mooring systems with suction-embedment " "anchors"
        ] = values

    # Data
    num_phases = len(plan)
    pos = np.arange(0.5, num_phases / 2.0 + 0.5, 0.5)

    ylabels = []
    customDates = []
    # for operation in Installation['OPERATION']:
    for operation in plan:
        l_phase = operation
        log_phase_descript = l_phase

        ylabels.append(log_phase_descript)
        start_dt = installation[l_phase]["Start date"] - timedelta(
            hours=installation[l_phase]["Prep time"]
        )
        prep_dt = installation[l_phase]["Start date"]
        depart_dt = installation[l_phase]["Depart date"]
        end_dt = installation[l_phase]["End date"]

        customDates.append(
            [
                date2num(start_dt),
                date2num(prep_dt),
                date2num(depart_dt),
                date2num(end_dt),
            ]
        )

    task_dates = {}

    for i, task in enumerate(ylabels):
        task_dates[task] = customDates[i]

    fig = plt.figure()
    ax = plt.subplot2grid((1, 2), (0, 1), colspan=1)

    # Plot the data:
    (
        start_date,
        end_prep_begin_wait_date,
        end_wait_begin_sea_date,
        end_date,
    ) = task_dates[ylabels[0]]

    ax.barh(
        0.5,
        (end_date - start_date),
        left=start_date,
        height=0.4,
        align="center",
        color="blue",
        alpha=0.75,
    )

    ax.barh(
        0.4,
        (end_prep_begin_wait_date - start_date),
        left=start_date,
        height=0.1,
        align="center",
        color="red",
        alpha=0.75,
        label="Prep Time",
    )

    ax.barh(
        0.5,
        (end_wait_begin_sea_date - end_prep_begin_wait_date),
        left=end_prep_begin_wait_date,
        height=0.1,
        align="center",
        color="yellow",
        alpha=0.75,
        label="Departure Delay",
    )

    ax.barh(
        0.6,
        (end_date - end_wait_begin_sea_date),
        left=end_wait_begin_sea_date,
        height=0.1,
        align="center",
        color="green",
        alpha=0.75,
        label="Sea Time",
    )

    for i in range(0, len(ylabels) - 1):
        (
            start_date,
            end_prep_begin_wait_date,
            end_wait_begin_sea_date,
            end_date,
        ) = task_dates[ylabels[i + 1]]

        ax.barh(
            (i * 0.5) + 1.0,
            (end_date - start_date),
            left=start_date,
            height=0.4,
            align="center",
            color="blue",
            alpha=0.75,
        )

        ax.barh(
            (i * 0.5) + 0.9,
            (end_prep_begin_wait_date - start_date),
            left=start_date,
            height=0.1,
            align="center",
            color="red",
            alpha=0.75,
        )

        ax.barh(
            (i * 0.5) + 1.0,
            (end_wait_begin_sea_date - end_prep_begin_wait_date),
            left=end_prep_begin_wait_date,
            height=0.1,
            align="center",
            color="yellow",
            alpha=0.75,
        )

        ax.barh(
            (i * 0.5) + 1.1,
            (end_date - end_wait_begin_sea_date),
            left=end_wait_begin_sea_date,
            height=0.1,
            align="center",
            color="green",
            alpha=0.75,
        )

    # Format the y-axis
    ylabels = ["\n".join(wrap(label, 40)) for label in ylabels]
    plt.yticks(pos, ylabels)

    # Format the x-axis
    ax.axis("tight")
    ax.set_ylim(ymin=-0.1, ymax=(num_phases) / 2 + 1.0)
    ax.grid(color="g", linestyle=":")

    ax.xaxis_date()  # Tell matplotlib that these are dates...

    rule = rrulewrapper(MONTHLY, interval=1)
    loc = RRuleLocator(rule)
    formatter = DateFormatter("%b '%y")

    ax.xaxis.set_major_locator(loc)
    ax.xaxis.set_major_formatter(formatter)

    for label in ax.get_xticklabels():
        label.set_rotation(30)

    # Format the legend
    font = font_manager.FontProperties(size="small")
    ax.legend(loc=0, prop=font)

    # Finish up
    ax.invert_yaxis()
    fig.autofmt_xdate()

    return fig


def installation_gantt_dates(dates, prep_time):
    gantt_dict = {
        "Start date": dates["Start"],
        "Depart date": dates["Depart"],
        "End date": dates["End"],
        "Prep time": prep_time,
    }

    return gantt_dict



# -*- coding: utf-8 -*-

#    Copyright (C) 2016 Mathew Topper, Vincenzo Nava
#    Copyright (C) 2017-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import logging
import math
import random
from typing import Optional

import pandas as pd

from .base import Strategy
from .basic import BasicStrategy

# Set up logging
module_logger = logging.getLogger(__name__)


class MultiSensitivity(Strategy):
    """A multi-variable sensitivity study  over a given range of
    values, adjusted before execution of a chosen module."""

    @property
    def version(self) -> int:
        """Version identifier for backwards compatibility when deserializing"""
        return 1

    @classmethod
    def get_name(cls):
        return "Multi Sensitivity"

    @classmethod
    def count_selections(cls, inputs_df, subsp_ratio):
        sorted_df = inputs_df.copy()

        # Create a multi-index & get selections
        sorted_df = sorted_df.set_index(["Module", "Variable"])
        selections = cls._get_selections(sorted_df, subsp_ratio)

        return len(selections)

    def configure(self, inputs_df, subspacing_ratio, skip_errors=True):
        config_dict = {
            "inputs_df": inputs_df,
            "subsp_ratio": subspacing_ratio,
            "skip_errors": skip_errors,
        }

        self.set_config(config_dict)

    def get_variables(self):
        if self._config is None:
            return None

        df = self._config["inputs_df"]
        result = df["Variable"].values

        return result

    def execute(self, core, project):
        # Test for Nones
        if (
            self._config is None
            or self._config["inputs_df"] is None
            or self._config["subsp_ratio"] is None
        ):
            errStr = (
                "The configuration values are None. Have you called "
                "the configure method?"
            )
            raise ValueError(errStr)

        inputs_df = self._config["inputs_df"]
        subsp_ratio = self._config["subsp_ratio"]

        # Check subspacing ratio
        if subsp_ratio is None or subsp_ratio > 1:
            subsp_ratio = 1.0
        elif subsp_ratio < 0:
            subsp_ratio = 0.0
        else:
            subsp_ratio = subsp_ratio

        # Sort the input frame and bulld the selection pool
        sorted_df = self._get_sorted_inputs(core, project, inputs_df)
        selections = self._get_selections(sorted_df, subsp_ratio)

        # Reset the index on the dataframe
        sorted_df = sorted_df.reset_index()

        # Get a branch to the first module to appear
        module_0 = sorted_df["Module"][0]
        mod_branch = self._tree.get_branch(core, project, module_0)

        # Check the project is active and record the simulation number
        sim_index = project.get_active_index()

        if sim_index is None:
            errStr = "Project has not been activated."
            raise RuntimeError(errStr)

        sim_keys = []
        sim_frames = []

        for i, selection_case in enumerate(selections[:-1]):
            # Create a dummy frame for building the simulations
            sim_df = sorted_df.copy()

            # Set the required values for the run
            sim_title = "Simulation {}".format(i)
            sim_df["Values"] = selection_case

            # Execute the simulation
            success_flag = self._safe_exe(core, project, sim_df, sim_title)

            # Create a new simulation clone and move to the required branch
            if success_flag:
                self.add_simulation_title(sim_title)
                sim_keys.append(sim_title)
                sim_frames.append(sim_df)

                core.clone_simulation(project)
                sim_index = project.get_active_index()

            mod_branch.reset(core, project)

        # Create a dummy frame for building the simulations
        sim_df = sorted_df.copy()

        # Complete the last simulation
        sim_title = "Simulation {}".format(i + 1)
        sim_df["Values"] = selections[-1]

        # Execute the simulation
        success_flag = self._safe_exe(core, project, sim_df, sim_title)

        if success_flag:
            self.add_simulation_title(sim_title)
            sim_keys.append(sim_title)
            sim_frames.append(sim_df)

        # Build the simulation details frame
        self.sim_details = pd.concat(sim_frames, keys=sim_keys)

    def _get_title_str(self, meta, value):
        title_str = "{} = {}".format(meta.title, value)

        if meta.units is not None:
            title_str = "{} ({})".format(title_str, meta.units[0])

        return title_str

    def _get_sorted_inputs(self, core, project, inputs_df):
        # Categorise module column by available modules
        list_modules = self._module_menu.get_available(core, project)

        sorted_df = inputs_df.copy()
        sorted_df["Module"] = pd.Categorical(sorted_df["Module"], list_modules)

        # Create a multi-index & sort
        sorted_df = sorted_df.set_index(["Module", "Variable"])
        sorted_df = sorted_df.sort_index(level=0)

        return sorted_df

    @classmethod
    def _get_selections(cls, sorted_df, subsp_ratio):
        # Generate simulation pool and selection
        values = sorted_df["Values"].tolist()
        pool = list(itertools.product(*values))

        pool_size = len(pool)
        number_samples = math.ceil(subsp_ratio * pool_size)

        random_selection = random.sample(
            range(0, pool_size), int(number_samples)
        )
        selections = [pool[x] for x in random_selection]

        return selections

    def _safe_exe(self, core, project, sim_df: pd.DataFrame, sim_title):
        success_flag = True

        assert self._config is not None
        if self._config["skip_errors"]:
            try:
                self._run_simulation(core, project, sim_df, sim_title)

            except (KeyboardInterrupt, SystemExit) as e:
                raise e

            except BaseException as e:
                msg = ("Passing exception '{}' for simulation " "{}").format(
                    type(e).__name__, sim_title
                )
                module_logger.exception(msg)

                success_flag = False

        else:
            # Execute the simulation
            self._run_simulation(core, project, sim_df, sim_title)

        return success_flag

    def _run_simulation(self, core, project, table_cs: pd.DataFrame, sim_title):
        # Set the simulation title
        project.set_simulation_title(sim_title)

        module_logger.info("Executing simulation '{}'".format(sim_title))

        # Set each variable
        for index, row in table_cs.iterrows():
            module = row["Module"]
            variable = row["Variable"]
            value = row["Values"]

            # Pick up the branch
            if module not in self._module_menu.get_available(core, project):
                errStr = "Module {} does not exist".format(module)
                raise ValueError(errStr)

            if module not in self._module_menu.get_active(core, project):
                errStr = "Module {} has not been activated".format(module)
                raise ValueError(errStr)

            mod_branch = self._tree.get_branch(core, project, module)

            # Check for existance of the variable
            module_inputs = mod_branch.get_input_status(core, project)

            msgStr = f"Variable {variable} is not an input to module {module}."
            if module_inputs is None:
                raise ValueError(msgStr)
            else:
                assert isinstance(module_inputs, dict)
                if variable not in module_inputs.keys():
                    raise ValueError(msgStr)

            multi_var = mod_branch.get_input_variable(core, project, variable)

            assert multi_var is not None
            multi_var.set_raw_interface(core, value)
            multi_var.read(core, project)

        # Borrow the Basic strategy
        basic = BasicStrategy()

        # Run the simulation
        basic.execute(core, project)

    @staticmethod
    def dump_config(config: Optional[dict]):
        if config is None:
            return None

        return {
            "inputs_df": config["inputs_df"].to_json(),
            "subsp_ratio": config["subsp_ratio"],
            "skip_errors": config["skip_errors"],
        }

    @staticmethod
    def load_config(serial_config: Optional[dict], version: int):
        if version != 1:
            raise RuntimeError("Data version not recognised")

        if serial_config is None:
            return None

        return {
            "inputs_df": pd.read_json(serial_config["inputs_df"]),
            "subsp_ratio": serial_config["subsp_ratio"],
            "skip_errors": serial_config["skip_errors"],
        }

    @staticmethod
    def dump_sim_details(sim_details: Optional[pd.DataFrame]):
        if sim_details is None:
            return None

        return sim_details.to_json()

    @staticmethod
    def load_sim_details(serial_sim_details: Optional[str]):
        if serial_sim_details is None:
            return None

        return pd.read_json(serial_sim_details)



import math


def bearing_to_radians(bearing):
    angle = 90.0 - bearing
    if angle <= -180.0:
        angle += 360.0

    angle = math.radians(angle)

    if angle < 0:
        angle += 2 * math.pi

    return angle


def radians_to_bearing(x):
    initial_bearing = 90 - math.degrees(x)
    compass_bearing = (initial_bearing + 360) % 360

    return compass_bearing



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime as dt
import json
import logging
import os
import platform
import re
import shutil
import time
from typing import Optional, Sequence

import numpy as np
import pandas as pd
import sqlalchemy
import yaml
from geoalchemy2 import Geometry
from mdo_engine.utilities.database import Database, PostGIS
from packaging.version import Version
from pandas.api.types import is_integer_dtype
from polite_config.configuration import ReadYAML
from polite_config.paths import ModPath, UserDataPath
from shapely import Point, get_srid, set_srid, wkt
from sqlalchemy.dialects import postgresql

from .files import onerror

# Set up logging
module_logger = logging.getLogger(__name__)

MIN_DB_VERSION = "2025.04.0"


def bathy_records_to_strata(bathy_records=None, pre_bathy=None):
    """Convert the bathymetry layers table returned by the database into
    Strata structure raw input"""

    loop_start_time = time.process_time()

    # Allow a predefined bathymetry table and grid dimensions to be passed
    # instead of the DB records.
    if bathy_records is None and pre_bathy is None:
        errStr = "One of arguments bathy_records or pre_bathy must be given"
        raise ValueError(errStr)

    elif bathy_records is not None:
        bathy_parts = init_bathy_records(bathy_records)
        if bathy_parts is None:
            module_logger.warning("No bathymetric points found")
            return None

        bathy_table, xi, yj = bathy_parts

    elif pre_bathy is not None:
        bathy_table, xi, yj = pre_bathy

    else:
        return None

    msg = "Building layers..."
    module_logger.debug(msg)

    layers = list(set(bathy_table["layer_order"]))
    layers.sort()

    bathy_layer_groups = bathy_table.groupby("layer_order")

    layer_depths = []
    layer_sediments = []

    for layer in layers:
        layer_table = bathy_layer_groups.get_group(layer)
        layer_depth, layer_sediment = build_bathy_layer(layer_table, xi, yj)

        layer_depths.append(layer_depth)
        layer_sediments.append(layer_sediment)

    depth_array = np.dstack(layer_depths)
    sediment_array = np.dstack(layer_sediments)

    layer_names = ["layer {}".format(x) for x in layers]

    raw_strata = {
        "values": {"depth": depth_array, "sediment": sediment_array},
        "coords": [xi, yj, layer_names],
    }

    loop_end_time = time.process_time()
    loop_time = loop_end_time - loop_start_time

    msg = ("Time elapsed building {} layer(s) was " "{} seconds").format(
        len(layers), loop_time
    )
    module_logger.debug(msg)

    return raw_strata


def tidal_series_records_to_xset(tidal_records):
    """Convert the bathymetry layers table returned by the database into
    tidal time series structure raw input"""

    loop_start_time = time.process_time()

    msg = "Building DataFrame from {} records".format(len(tidal_records))
    module_logger.debug(msg)

    tidal_table = pd.DataFrame.from_records(
        tidal_records,
        columns=[
            "utm_point",
            "measure_date",
            "measure_time",
            "u",
            "v",
            "turbulence_intensity",
            "ssh",
        ],
    )

    if tidal_table.empty:
        return None

    msg = "Converting PostGIS Point types to coordinates..."
    module_logger.debug(msg)

    tidal_table = point_to_xy(tidal_table)

    msg = "Getting grid extents..."
    module_logger.debug(msg)

    xi, yj = get_grid_coords(tidal_table)

    msg = "Joining dates and times..."
    module_logger.debug(msg)

    tidal_table["datetime"] = [
        dt.datetime.combine(d, t)
        for d, t in zip(
            tidal_table["measure_date"], tidal_table["measure_time"]
        )
    ]

    tidal_table = tidal_table.drop("measure_date", axis=1)
    assert tidal_table is not None
    tidal_table = tidal_table.drop("measure_time", axis=1)
    assert tidal_table is not None

    msg = "Building time steps..."
    module_logger.debug(msg)

    steps = list(set(tidal_table["datetime"]))
    steps.sort()

    tidal_table_groups = tidal_table.groupby("datetime")

    u_steps = []
    v_steps = []
    ssh_steps = []
    ti_steps = []

    for step in steps:
        step_table = tidal_table_groups.get_group(step)
        (u_step, v_step, ssh_step, ti_step) = build_tidal_series_step(
            step_table, xi, yj
        )

        u_steps.append(u_step)
        v_steps.append(v_step)
        ssh_steps.append(ssh_step)
        ti_steps.append(ti_step)

    u_array = np.dstack(u_steps)
    v_array = np.dstack(v_steps)
    ssh_array = np.dstack(ssh_steps)
    ti_array = np.dstack(ti_steps)

    raw = {
        "values": {
            "U": u_array,
            "V": v_array,
            "SSH": ssh_array,
            "TI": ti_array,
        },
        "coords": [xi, yj, steps],
    }

    loop_end_time = time.process_time()
    loop_time = loop_end_time - loop_start_time

    msg = ("Time elapsed building {} step(s) was " "{} seconds").format(
        len(steps), loop_time
    )
    module_logger.debug(msg)

    return raw


def init_bathy_records(bathy_records):
    msg = "Building DataFrame from {} records".format(len(bathy_records))
    module_logger.debug(msg)

    bathy_cols = ["utm_point", "depth"]
    bathy_cols.extend(["layer_order", "initial_depth", "sediment_type"])
    bathy_table = pd.DataFrame.from_records(bathy_records, columns=bathy_cols)

    if bathy_table.empty:
        return None

    msg = "Converting PostGIS Point types to coordinates..."
    module_logger.debug(msg)

    bathy_table = point_to_xy(bathy_table)

    msg = "Getting grid extents..."
    module_logger.debug(msg)

    xi, yj = get_grid_coords(bathy_table)

    return bathy_table, xi, yj


def point_to_xy(
    df: pd.DataFrame,
    point_column="utm_point",
    decimals=2,
    drop_point_column=True,
):
    x = []
    y = []

    for point in df[point_column]:
        assert isinstance(point, Point)
        coords = list(point.coords)[0]
        x.append(coords[0])
        y.append(coords[1])

    x = np.array(x)
    x = x.round(decimals)

    y = np.array(y)
    y = y.round(decimals)

    df["x"] = x
    df["y"] = y

    if drop_point_column:
        df_dropped = df.drop(point_column, axis=1)
        assert df_dropped is not None
        df = df_dropped

    return df


def get_grid_coords(df: pd.DataFrame, xlabel="x", ylabel="y"):
    xi = np.unique(df[xlabel])

    ## TODO: Find a better method to check for valid grid spacing
    # for x in xi:
    #     yx: pd.Series = df[ylabel][df[xlabel] == x].sort_values().unique()
    #     if len(yx) < 2:
    #         continue

    #     ydist = np.diff(yx)

    #     if len(np.unique(ydist)) != 1:
    #         safe_dist = [str(y) for y in np.unique(ydist)]
    #         dist_str = ", ".join(safe_dist)
    #         errStr = (
    #             f"Distances in y-direction for x = {x} are not equal. Unique "
    #             f"lengths are: {dist_str}"
    #         )
    #         raise ValueError(errStr)

    yj = np.unique(df[ylabel])

    ## TODO: Find a better method to check for valid grid spacing
    # for y in yj:
    #     xy: pd.Series = df[xlabel][df[ylabel] == y].sort_values().unique()
    #     if len(xy) < 2:
    #         continue

    #     xdist = np.diff(xy)

    #     if len(np.unique(xdist)) != 1:
    #         safe_dist = [str(x) for x in np.unique(xdist)]
    #         dist_str = ", ".join(safe_dist)
    #         errStr = (
    #             f"Distances in x-direction for y = {y} are not equal. Unique "
    #             f"lengths are: {dist_str}"
    #         )
    #         raise ValueError(errStr)

    return xi, yj


def build_bathy_layer(layer_table, xi, yj):
    depth_array = np.zeros([len(xi), len(yj)]) * np.nan
    sediment_array = np.full([len(xi), len(yj)], None, dtype="object")

    for record in layer_table.itertuples():
        xidxs = np.where(xi == record.x)[0]
        assert len(xidxs) == 1
        xidx = xidxs[0]

        yidxs = np.where(yj == record.y)[0]
        assert len(yidxs) == 1
        yidx = yidxs[0]

        depth = record.depth - record.initial_depth
        sediment = record.sediment_type

        depth_array[xidx, yidx] = depth
        sediment_array[xidx, yidx] = sediment

    return depth_array, sediment_array


def build_tidal_series_step(step_table, xi, yj):
    u_array = np.zeros([len(xi), len(yj)]) * np.nan
    v_array = np.zeros([len(xi), len(yj)]) * np.nan
    ssh_array = np.zeros([len(xi), len(yj)]) * np.nan
    ti_array = np.zeros([len(xi), len(yj)]) * np.nan

    for record in step_table.itertuples():
        xidxs = np.where(xi == record.x)[0]
        assert len(xidxs) == 1
        xidx = xidxs[0]

        yidxs = np.where(yj == record.y)[0]
        assert len(yidxs) == 1
        yidx = yidxs[0]

        u_array[xidx, yidx] = record.u
        v_array[xidx, yidx] = record.v
        ssh_array[xidx, yidx] = record.ssh
        ti_array[xidx, yidx] = record.turbulence_intensity

    return u_array, v_array, ssh_array, ti_array


def get_table_df(db, schema, table, columns):
    df = pd.read_sql_table(table, db._engine, schema=schema, columns=columns)
    return df


def get_one_from_column(db, schema, table, column):
    Table = db.safe_reflect_table(table, schema)
    result = db.session.query(Table.columns[column]).one_or_none()

    return result


def filter_one_from_column(
    db, schema, table, result_column, filter_column, filter_value
):
    TableTwo = db.safe_reflect_table(table, schema)
    query = db.session.query(TableTwo.columns[result_column])
    result = query.filter(
        TableTwo.columns[filter_column] == filter_value
    ).one_or_none()

    return result


def get_all_from_columns(db, schema, table, columns):
    Table = db.safe_reflect_table(table, schema)

    col_lists = []

    for column in columns:
        col_all = db.session.query(Table.columns[column]).all()
        trim_col = [item[0] for item in col_all]

        col_lists.append(trim_col)

    return col_lists


def database_to_files(
    root_path,
    table_list,
    database,
    schema=None,
    table_name_list=None,
    pid_list=None,
    join_list=None,
    where_list=None,
    auto_child=False,
    prefer_csv=False,
    print_function=None,
):
    def _dump_child(
        root_path,
        table_dict,
        engine,
        schema,
        table_name_list=None,
        pid_list=None,
        fid_list=None,
        where_list=None,
        auto_val=None,
    ):
        child_path = os.path.join(root_path, table_dict["table"])
        if auto_val is not None:
            child_path += str(auto_val)
            auto_child = True
        else:
            auto_child = False

        # dump a directory
        if os.path.exists(child_path):
            shutil.rmtree(child_path, onexc=onerror)

        os.makedirs(child_path)

        # Recurse for the children
        database_to_files(
            child_path,
            table_dict["children"],
            engine,
            schema,
            table_name_list,
            pid_list,
            fid_list,
            where_list,
            auto_child,
            prefer_csv,
            print_function,
        )

    def _autofit_columns(xlpath):
        if platform.system() != "Windows":
            return

        from win32com.client import Dispatch  # type: ignore

        excel = Dispatch("Excel.Application")
        wb = excel.Workbooks.Open(os.path.abspath(xlpath))

        excel.Worksheets(1).Activate()
        excel.ActiveSheet.Columns.AutoFit()

        wb.Save()
        wb.Close()

    if print_function is None:

        def _print_function(x: str):
            pass

        print_function = _print_function

    # Clean and setup dump directory
    if os.path.exists(root_path):
        shutil.rmtree(root_path, onexc=onerror)

    os.makedirs(root_path)

    for table_dict in table_list:
        table_df = None
        new_name_list = None
        new_pid_list = None
        new_join_list = None
        new_where_list = None

        full_dict = check_dict(table_dict)

        # Set the schema
        if schema is None:
            var_schema = full_dict["schema"]
        else:
            var_schema = schema

        msg_str = "Dumping table: {}.{}".format(var_schema, table_dict["table"])
        print_function(msg_str)

        if not full_dict["dummy"]:
            if table_name_list is None:
                table_idx = 0
                new_name_list = [full_dict["table"]]

            else:
                table_idx = len(table_name_list)
                new_name_list = table_name_list + [full_dict["table"]]

                join = full_dict["join"]
                if join is None:
                    raise RuntimeError("join must be defined for child tables")

                if join_list is None:
                    new_join_list = [join]
                else:
                    new_join_list = join_list + [join]

            # Filter by the parent table if required
            query_str = query_builder(
                new_name_list,
                pid_list,
                new_join_list,
                where_list,
                var_schema,
            )

            msg_str = "Executing query: {}".format(query_str)
            print_function(msg_str)

            # Read the table first
            table_df = pd.read_sql(query_str, database._engine)

            if full_dict["stripf"] and auto_child:
                msg_str = "Stripping column: {}".format(full_dict["join"])
                print_function(msg_str)

                table_df = table_df.drop(full_dict["join"], axis=1)
                assert table_df is not None

            if full_dict["array"] is not None:
                array_str = ", ".join(full_dict["array"])
                msg_str = "Coverting array columns: {}".format(array_str)
                print_function(msg_str)

                table_df = convert_array(table_df, full_dict["array"])

            if full_dict["bool"] is not None:
                bool_str = ", ".join(full_dict["bool"])
                msg_str = "Coverting boolean columns: {}".format(bool_str)
                print_function(msg_str)

                table_df = convert_bool(table_df, full_dict["bool"])

            if full_dict["geo"] is not None:
                geo_str = ", ".join(full_dict["geo"])
                msg_str = "Coverting Geometry columns: {}".format(geo_str)
                print_function(msg_str)

                table_df = convert_geo(table_df, full_dict["geo"])

            if full_dict["time"] is not None:
                time_str = ", ".join(full_dict["time"])
                msg_str = "Coverting Time columns: {}".format(time_str)
                print_function(msg_str)

                table_df = convert_time(table_df, full_dict["time"])

            table_df.sort_values(by=[full_dict["pkey"]], inplace=True)

            if len(table_df) < 1e6 and os.name == "nt" and not prefer_csv:
                table_fname = full_dict["table"] + ".xlsx"
                tab_path = os.path.join(root_path, table_fname)

                msg_str = "Writing to: {}".format(tab_path)
                print_function(msg_str)

                # Create a Pandas Excel writer using XlsxWriter as the engine.
                writer = pd.ExcelWriter(tab_path, engine="openpyxl")

                # Convert the dataframe to an XlsxWriter Excel object.
                table_df.to_excel(writer, index=False)

                # Close the Pandas Excel writer and output the Excel file.
                writer.close()

                # Fit the column widths (don't let failure be catastrophic)
                try:
                    _autofit_columns(tab_path)
                except Exception:
                    print_function("*** Column adjust failed. Skipping. ***")
                    pass

            else:
                table_fname = full_dict["table"] + ".csv"
                tab_path = os.path.join(root_path, table_fname)

                msg_str = "Writing to: {}".format(tab_path)
                print_function(msg_str)

                table_df.to_csv(tab_path, index=False)

        if full_dict["children"] is not None:
            # Include pid in iteration
            if not full_dict["dummy"]:
                pkey = full_dict["pkey"]

                if pid_list is None:
                    new_pid_list = [pkey]
                else:
                    new_pid_list = pid_list + [pkey]

            # Check autokey
            if full_dict["autokey"]:
                assert table_df is not None
                pkids = table_df[full_dict["pkey"]]
                del table_df

                for pkid in pkids:
                    # Add a where
                    new_where = {"table#": table_idx, "value": pkid}

                    if where_list is None:
                        new_where_list = [new_where]
                    else:
                        new_where_list = where_list + [new_where]

                    _dump_child(
                        root_path,
                        table_dict,
                        database,
                        var_schema,
                        new_name_list,
                        new_pid_list,
                        new_join_list,
                        new_where_list,
                        pkid,
                    )

            else:
                del table_df

                if where_list is None:
                    new_where_list = None
                else:
                    new_where_list = where_list[:]

                _dump_child(
                    root_path,
                    table_dict,
                    database,
                    var_schema,
                    new_name_list,
                    new_pid_list,
                    new_join_list,
                    new_where_list,
                )


def database_from_files(
    root_path,
    table_list,
    database,
    schema=None,
    add_fid=None,
    truncate=True,
    drop_missing=True,
    parent: Optional[str] = None,
    first_idx_map: Optional[dict[str, int]] = None,
    offset_maps: Optional[dict[str, dict[int, int]]] = None,
    print_function=None,
):
    if print_function is None:

        def _print_function(x: str):
            pass

        print_function = _print_function

    def _list_dirs(path, tab_match):
        dir_list = [
            item
            for item in os.listdir(path)
            if os.path.isdir(os.path.join(path, item))
        ]

        match_str = r"^{}[0-9]+".format(tab_match)
        dir_list = [s for s in dir_list if re.search(match_str, s)]
        dir_list.sort()

        return dir_list

    for table_dict in table_list:
        full_dict = check_dict(table_dict)

        # Set the schema
        if schema is None:
            var_schema = full_dict["schema"]
        else:
            var_schema = schema

        if not full_dict["dummy"]:
            xlname = "{}.xlsx".format(full_dict["table"])
            xlpath = os.path.join(root_path, xlname)

            csvname = "{}.csv".format(full_dict["table"])
            csvpath = os.path.join(root_path, csvname)

            # Try to read the table as xl or csv
            if os.path.isfile(xlpath):
                msg_str = "Reading file: {}".format(xlpath)
                print_function(msg_str)

                xl = pd.ExcelFile(xlpath)
                df = xl.parse("Sheet1")

            elif os.path.isfile(csvpath):
                msg_str = "Reading file: {}".format(csvpath)
                print_function(msg_str)

                df = pd.read_csv(csvpath)

            else:
                errStr = (
                    "Table {} could not be found in directory " "{}"
                ).format(full_dict["table"], root_path)
                raise IOError(errStr)

            if add_fid is not None:
                msg_str = ("Adding foreign key '{}' with value: " "{}").format(
                    full_dict["join"], add_fid
                )
                print_function(msg_str)

                df[full_dict["join"]] = add_fid

            # Get the table name
            dbname = "{}.{}".format(var_schema, full_dict["table"])

            # Clean the table
            if truncate:
                msg_str = "Truncating table: {}".format(dbname)
                print_function(msg_str)

                query_str = "TRUNCATE TABLE {} CASCADE".format(dbname)
                database.execute_transaction(query_str)

            # Drop columns not in the recepting table
            if drop_missing:
                actual_tables = database.get_column_names(
                    full_dict["table"], var_schema
                )
                missing_set = set(df.columns) - set(actual_tables)

                if missing_set:
                    cols_str = ", ".join(missing_set)
                    msg_str = ("Dropping extraneous columns: " "{}").format(
                        cols_str
                    )
                    print_function(msg_str)

                    df = df.drop(list(missing_set), axis=1)
                    assert df is not None

            dtype = {}

            if full_dict["bool"] is not None:
                bool_str = ", ".join(full_dict["bool"])
                msg_str = "Reverting boolean columns: {}".format(bool_str)
                print_function(msg_str)

                df = revert_bool(df, full_dict["bool"])

            if full_dict["geo"] is not None:
                geo_str = ", ".join(full_dict["geo"])
                msg_str = "Reverting Geometry columns: {}".format(geo_str)
                print_function(msg_str)

                df = revert_geo(df, full_dict["geo"])

                for x in full_dict["geo"]:
                    dtype[x] = Geometry

            if full_dict["array"] is not None:
                array_str = ", ".join(full_dict["array"])
                msg_str = "Reverting array columns: {}".format(array_str)
                print_function(msg_str)

                df = revert_array(df, full_dict["array"])

                for x in full_dict["array"]:
                    dtype[x] = postgresql.ARRAY(sqlalchemy.types.REAL)

            if full_dict["time"] is not None:
                for x in full_dict["time"]:
                    dtype[x] = postgresql.TIME

            if full_dict["date"] is not None:
                for x in full_dict["date"]:
                    dtype[x] = postgresql.DATE

            df.sort_values(by=[full_dict["pkey"]], inplace=True)

            if is_integer_dtype(df[full_dict["pkey"]]):
                msg_str = "Resetting primary and foreign keys"
                print_function(msg_str)

                if first_idx_map is None:
                    first_idx = 1
                elif dbname in first_idx_map:
                    first_idx = first_idx_map[dbname]
                else:
                    first_idx = 1

                offset_map = get_offset_map(df, full_dict["pkey"], first_idx)
                to_replace = {full_dict["pkey"]: offset_map}

                if full_dict["fkeys"] is not None:
                    fkeys = full_dict["fkeys"]
                else:
                    fkeys = {}

                if full_dict["join"] is not None:
                    if parent is None:
                        raise ValueError(
                            "parent must not be None for joined tables"
                        )

                    fkeys[full_dict["join"]] = parent

                if fkeys:
                    if offset_maps is None:
                        raise ValueError(
                            "offset_maps must not be None when join or fkeys "
                            "are defined"
                        )
                    for fkey, ptable in fkeys.items():
                        to_replace[fkey] = offset_maps[ptable]

                for key, val_map in to_replace.items():
                    df[key] = df[key].map(val_map)

            msg_str = "Writing to table: {}".format(dbname)
            print_function(msg_str)

            if not dtype:
                dtype = None

            df.to_sql(
                full_dict["table"],
                database._engine,
                schema=var_schema,
                if_exists="append",
                index=False,
                chunksize=50000,
                dtype=dtype,
            )

            # Record tables next available index and pkey offset
            if is_integer_dtype(df[full_dict["pkey"]]):
                next_idx = df[full_dict["pkey"]].max() + 1

                if first_idx_map is None:
                    first_idx_map = {dbname: next_idx}
                else:
                    first_idx_map[dbname] = next_idx

                if offset_maps is None:
                    offset_maps = {dbname: offset_map}
                elif dbname not in offset_maps:
                    offset_maps[dbname] = offset_map
                else:
                    offset_maps[dbname].update(offset_map)

            del df

        if full_dict["children"] is not None:
            if full_dict["autokey"]:
                tab_dirs = _list_dirs(root_path, full_dict["table"])
                fids = [int(x.split(full_dict["table"])[1]) for x in tab_dirs]

                if not tab_dirs:
                    continue

                # Truncate table for first ID but not others
                first_dir = tab_dirs.pop(0)
                first_fid = fids.pop(0)

                child_path = os.path.join(root_path, first_dir)

                first_idx_map, offset_maps = database_from_files(
                    child_path,
                    full_dict["children"],
                    database,
                    var_schema,
                    first_fid,
                    parent=dbname,
                    first_idx_map=first_idx_map,
                    offset_maps=offset_maps,
                    print_function=print_function,
                )

                for next_tab_dir, next_fid in zip(tab_dirs, fids):
                    child_path = os.path.join(root_path, next_tab_dir)

                    first_idx_map, offset_maps = database_from_files(
                        child_path,
                        full_dict["children"],
                        database,
                        var_schema,
                        next_fid,
                        False,
                        parent=dbname,
                        first_idx_map=first_idx_map,
                        offset_maps=offset_maps,
                        print_function=print_function,
                    )

            else:
                child_path = os.path.join(root_path, full_dict["table"])

                database_from_files(
                    child_path,
                    full_dict["children"],
                    database,
                    var_schema,
                    truncate=truncate,
                    parent=dbname,
                    first_idx_map=first_idx_map,
                    offset_maps=offset_maps,
                    print_function=print_function,
                )

    return first_idx_map, offset_maps


def query_builder(
    table_list,
    pid_list=None,
    join_list=None,
    where_list=None,
    schema=None,
):
    def _add_schema(table_name, schema=None):
        if schema is None:
            dbname = table_name
        else:
            dbname = "{}.{}".format(schema, table_name)

        return dbname

    consume_list = table_list[:]
    table_name = _add_schema(consume_list.pop(), schema)

    # No joins or wheres
    if pid_list is None:
        query_str = "SELECT * FROM {};".format(table_name)
        return query_str

    table_shorts = ["t{}".format(i) for i in range(len(table_list))]
    consume_shorts = table_shorts[:]
    table_short = consume_shorts.pop()

    query_str = "SELECT {0}.*\nFROM {1} {0}".format(table_short, table_name)

    # Add joins
    if join_list is not None:
        consume_fid = join_list[:]
        consume_pid = pid_list[:]

        while consume_list:
            table_fid = consume_fid.pop()
            join_table_pid = consume_pid.pop()
            join_table_name = _add_schema(consume_list.pop(), schema)
            join_table_short = consume_shorts.pop()

            query_str += ("\nJOIN {0} {1} ON {1}.{2} = {3}.{4}").format(
                join_table_name,
                join_table_short,
                join_table_pid,
                table_short,
                table_fid,
            )

            table_short = join_table_short

    # Add wheres
    if where_list is not None:
        where_str = None

        for where_dict in where_list:
            table_short = table_shorts[where_dict["table#"]]
            table_pid = pid_list[where_dict["table#"]]
            pid_value = where_dict["value"]

            eq_str = "{}.{} = {}".format(table_short, table_pid, pid_value)

            if where_str is None:
                where_str = "\nWHERE " + eq_str
            else:
                where_str += " AND " + eq_str

        if where_str is not None:
            query_str += where_str

    query_str += ";"

    return query_str


def convert_array(table_df, array_cols):
    brackets = str.maketrans("[]", "{}")

    def _safe_square2curly(x):
        if not isinstance(x, list) and pd.isna(x):
            return
        else:
            y = str(x).translate(brackets)
            return y

    for array_col in array_cols:
        table_df[array_col] = table_df[array_col].apply(_safe_square2curly)

    return table_df


def revert_array(table_df, array_cols):
    brackets = str.maketrans("{}", "[]")

    def _safe_curly2list(x):
        if pd.isna(x):
            return
        else:
            y = eval(x.translate(brackets))
            return y

    for array_col in array_cols:
        table_df[array_col] = table_df[array_col].apply(_safe_curly2list)

    return table_df


def convert_bool(table_df, bool_cols):
    def _safe_bool2str(x):
        if pd.isna(x):
            y = None
        elif x:
            y = "yes"
        else:
            y = "no"

        return y

    for bool_col in bool_cols:
        table_df[bool_col] = table_df[bool_col].apply(_safe_bool2str)

    return table_df


def revert_bool(table_df, bool_cols):
    def _safe_str2bool(x):
        if pd.isna(x):
            y = None
        elif isinstance(x, bool):
            y = x
        elif x == "yes":
            y = True
        elif x == "no":
            y = False
        else:
            y = None

        return y

    for bool_col in bool_cols:
        table_df[bool_col] = table_df[bool_col].apply(_safe_str2bool)

    return table_df


def convert_geo(table_df, geo_cols):
    def _safe_to_wkt(x):
        if pd.isna(x):
            return
        else:
            srid = get_srid(x)
            if srid > 0:
                result = "SRID={};{}".format(srid, x.wkt)
            else:
                result = x.wkt
            return result

    for geo_col in geo_cols:
        table_df[geo_col] = table_df[geo_col].apply(_safe_to_wkt)

    return table_df


def revert_geo(table_df: pd.DataFrame, geo_cols: Sequence[str]):
    def _safe_to_wkb(x):
        if pd.isna(x):
            return
        else:
            srid = None
            parts = x.split(";")
            if len(parts) == 2:
                srid = int(parts[0][5:])
                geom = parts[1]
            else:
                geom = parts[0]

            geo_shape = wkt.loads(geom)
            if srid is not None:
                geo_shape = set_srid(geo_shape, srid)

            return geo_shape

    for geo_col in geo_cols:
        table_df[geo_col] = table_df[geo_col].apply(_safe_to_wkb)  # type: ignore

    return table_df


def convert_time(table_df, time_cols):
    def _safe_time2str(x):
        if pd.isna(x):
            return
        else:
            return x.strftime("%H:%M:%S")

    for time_col in time_cols:
        table_df[time_col] = table_df[time_col].apply(_safe_time2str)

    return table_df


def check_dict(table_dict):
    full_dict = {
        "array": None,
        "autokey": False,
        "bool": None,
        "children": None,
        "dummy": False,
        "join": None,
        "geo": None,
        "pkey": "id",
        "fkeys": None,
        "schema": None,
        "stripf": False,
        "time": None,
        "date": None,
    }

    full_dict.update(table_dict)

    if "table" not in full_dict.keys():
        errStr = (
            "Each definition requires a table name under the " "'table' key."
        )
        raise KeyError(errStr)

    return full_dict


def get_offset_map(
    df: pd.DataFrame,
    column_name: str,
    first: int,
) -> dict[int, int]:
    ordered_df = df.reset_index(drop=True)
    value = pd.to_numeric(ordered_df.at[0, column_name], downcast="integer")
    column: pd.Series[int] = df[column_name]
    offset = first - value
    new = column + offset
    offset_map: dict[int, int] = {
        x: y for x, y in zip(column.values, new.values)
    }

    return offset_map


def get_table_map(map_name="table_map.yaml"):
    # Load the yaml files
    objdir = ModPath(__name__, "..", "config")
    table_yaml = objdir / map_name

    with open(table_yaml, "r") as f:
        table_list = yaml.load(f, Loader=yaml.FullLoader)

    return table_list


def filter_map(table_list, filter_name, parent=None):
    copy_list = table_list[:]

    for table_dict in copy_list:
        full_dict = check_dict(table_dict)

        table_name = full_dict["table"]

        if filter_name == table_name:
            if parent is not None:
                parent["children"] = [full_dict]
                return parent
            else:
                return table_dict

        elif full_dict["children"] is not None:
            result = filter_map(full_dict["children"], filter_name, table_dict)

            if result is not None:
                if parent is not None:
                    parent["children"] = [result]
                    return parent
                else:
                    return result

    return None


def draw_map(table_list, level=0):
    map_str = ""

    for table_dict in table_list:
        full_dict = check_dict(table_dict)

        if level > 0:
            n_spaces = 2 * level - 1
            level_marks = " " + " " * n_spaces + "|"
        else:
            level_marks = "|"

        level_marks += "-"

        map_str += "{} {}\n".format(level_marks, full_dict["table"])

        if full_dict["children"] is not None:
            map_str += draw_map(full_dict["children"], level + 1)

    return map_str


def get_database_config(db_config_name="database.yaml"):
    userconfigdir = UserDataPath("dtocean_core", "DTOcean", "config")
    useryaml = ReadYAML(userconfigdir, db_config_name)

    if (userconfigdir / db_config_name).is_file():
        configdir = userconfigdir
    else:
        configdir = ModPath("dtocean_core", "config")

    configyaml = ReadYAML(configdir, db_config_name)
    config = configyaml.read()

    return useryaml, config


def get_database(
    credentials,
    echo=False,
    timeout=None,
    db_adapter="psycopg",
    min_version: Optional[str] = None,
):
    database = PostGIS(db_adapter)
    database.set_credentials(credentials)
    database.set_echo(echo)
    database.set_timeout(timeout)
    database.configure()

    if min_version is not None:
        actual_version = get_database_version(database)
        if actual_version is None:
            raise RuntimeError("Database version is unknown")

        if Version(actual_version) < Version(min_version):
            raise RuntimeError(
                f"Database version {actual_version} is less than the required "
                f"{min_version}"
            )

    return database


def get_database_version(database: Database) -> Optional[str]:
    credentials = database.get_credentials()
    query = f"""SELECT pg_catalog.shobj_description(d.oid, 'pg_database') AS "Description"
FROM pg_catalog.pg_database d
WHERE datname = '{credentials["dbname"]}';"""
    with database.exectute_query(query) as cur:
        first = cur.fetchone()
        if first is None:
            return None
        meta_str = first[0]

    if not meta_str:
        return None

    try:
        meta = json.loads(meta_str)
    except (json.JSONDecodeError, TypeError):
        return None

    if "version" not in meta:
        return None

    return meta["version"]



# -*- coding: utf-8 -*-

#    Copyright (C) 2021-2026 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import Strategy

# Set up logging
module_logger = logging.getLogger(__name__)


class BasicStrategy(Strategy):
    """A basic strategy which will run all selected modules and themes in
    sequence."""

    @property
    def version(self) -> int:
        """Version identifier for backwards compatibility when deserializing"""
        return 1

    @classmethod
    def get_name(cls):
        return "Basic"

    def configure(self, kwargs=None):
        """Does nothing in this case"""

    def get_variables(self):
        return None

    def execute(self, core, project):
        # Check the project is active and record the simulation title
        sim_index = project.get_active_index()

        if sim_index is None:
            errStr = "Project has not been activated."
            raise RuntimeError(errStr)

        sim_title = project.get_simulation_title(index=sim_index)
        self.add_simulation_title(sim_title)

        current_mod = self._module_menu.get_current(core, project)

        while current_mod:
            self._module_menu.execute_current(
                core, project, allow_unavailable=True
            )
            current_mod = self._module_menu.get_current(core, project)

    @staticmethod
    def dump_config(config):
        return None

    @staticmethod
    def load_config(serial_config, version):
        return None

    @staticmethod
    def dump_sim_details(sim_details):
        return None

    @staticmethod
    def load_sim_details(serial_sim_details):
        return None



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Wed Apr 06 15:59:04 2016

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize

from .base import PlotInterface


class WaveOccurrencePlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Wave Resource Occurrence Matrix"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["farm.wave_occurrence"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {"occurrence_matrix": "farm.wave_occurrence"}

        return id_map

    def connect(self):
        occurrence_grid = self.data.occurrence_matrix
        occurrence_flat = occurrence_grid.sum("Dir")
        vals = occurrence_flat.values.T

        col_normals = Normalize(vals.min(), vals.max())

        cellStrs = []

        for row in vals:
            rowStrs = ["{:.2%}".format(x) for x in row]
            cellStrs.append(rowStrs)

        fig = plt.figure()
        fig.add_subplot(111, frameon=True, xticks=[], yticks=[])

        # Add headers and a table at the bottom of the axes
        plt.table(
            cellText=[[""]],
            colLabels=["Te [$s$]"],
            loc="center",
            bbox=[0, 0.85, 1.0, 0.1],
        )

        plt.table(
            cellText=[["Bob"]],
            rowLabels=["Hm0 [$m$]          "],
            loc="center",
            bbox=[0, 0, 1, 0.845],
        )

        plt.table(
            cellText=cellStrs,
            rowLabels=occurrence_flat.Hm0.values,
            colLabels=occurrence_flat.Te.values,
            cellColours=mpl.colormaps["RdYlGn_r"](col_normals(vals)),
            loc="center",
            bbox=[0, 0, 1, 0.9],
        )

        plt.title("Joint Probability Plot (Over All Directions)")

        self.fig_handle = plt.gcf()


class PowerMatrixPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Single Wave Device Power Matrix"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["device.wave_power_matrix"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {"power_matrix": "device.wave_power_matrix"}

        return id_map

    def connect(self):
        occurrence_grid = self.data.power_matrix
        occurrence_flat = occurrence_grid.mean("Dir")
        vals = occurrence_flat.values.T

        col_normals = Normalize(vals.min(), vals.max())

        cellStrs = []

        for row in vals:
            rowStrs = ["{0:d}".format(int(x)) for x in row]
            cellStrs.append(rowStrs)

        fig = plt.figure()
        fig.add_subplot(111, frameon=True, xticks=[], yticks=[])

        # Add headers and a table at the bottom of the axes
        plt.table(
            cellText=[[""]],
            colLabels=["Te [$s$]"],
            loc="center",
            bbox=[0, 0.85, 1.0, 0.1],
        )

        plt.table(
            cellText=[["Bob"]],
            rowLabels=["Hm0 [$m$]          "],
            loc="center",
            bbox=[0, 0, 1, 0.8575],
        )

        short_hm0 = ["{:.2f}".format(x) for x in occurrence_flat.Hm0.values]
        short_te = ["{:.2f}".format(x) for x in occurrence_flat.Te.values]

        plt.table(
            cellText=cellStrs,
            rowLabels=short_hm0,
            colLabels=short_te,
            cellColours=mpl.colormaps["RdYlGn_r"](col_normals(vals)),
            loc="center",
            bbox=[0, 0, 1, 0.9],
        )

        plt.title("Single Device Power Matrix (Mean Over All Directions) [kW]")

        self.fig_handle = plt.gcf()


class TeHm0Plot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Te & Hm0 Time Series"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["farm.wave_series"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {"wave_series": "farm.wave_series"}

        return id_map

    def connect(self):
        fig = plt.figure()
        ax1 = fig.add_subplot(111)

        # Add headers and a table at the bottom of the axes
        ax1 = self.data.wave_series.Te.plot(ax=ax1)
        ax2 = self.data.wave_series.Hm0.plot(secondary_y=True, style="g")

        ax1.set_ylabel("Te [$s$]")
        ax2.set_ylabel("Hm0 [$m$]")

        plt.title("Te & Hm0 Time Series")

        self.fig_handle = plt.gcf()



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2026 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Fri Jul 29 17:44:49 2016

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

from mdo_engine.boundary.interface import QueryInterface
from shapely import Polygon, to_wkt

from dtocean_core.utils.database import (
    bathy_records_to_strata,
    tidal_series_records_to_xset,
)


class LeaseBathyInterface(QueryInterface):
    """Interface to filter the database for a selected site and technology."""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Lease Area Bathymetry Filtering Interface"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["site.lease_boundary"]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = ["bathymetry.layers"]

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example:

              optional = ["My:first:variable",
                         ]
        """

        option_list = None

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "bathymetry": "bathymetry.layers",
            "lease_poly": "site.lease_boundary",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """

        if self._db is None:
            raise RuntimeError("No database connected")

        # Manipulate wkt string for function requirements
        lease_poly: Polygon = self.data.lease_poly
        poly_wkt = to_wkt(lease_poly)
        func_poly_str = poly_wkt.replace("POLYGON ((", "")[:-2]

        query_str = (
            "SELECT "
            "utm_point, "
            "depth, "
            "layer_order, "
            "initial_depth, "
            "sediment_type "
            "FROM "
            "filter.sp_select_bathymetry_by_polygon"
            "('{}')"
        ).format(func_poly_str)

        result = self._db.server_execute_query(query_str)
        raw_strata = bathy_records_to_strata(result)

        self.data.bathymetry = raw_strata


class CorridorBathyInterface(QueryInterface):
    """Interface to filter the database for a selected site and technology."""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Cable Corridor Bathymetry Filtering Interface"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["site.corridor_boundary"]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = ["corridor.layers"]

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example:

              optional = ["My:first:variable",
                         ]
        """

        option_list = None

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "corridor": "corridor.layers",
            "corridor_poly": "site.corridor_boundary",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """
        if self._db is None:
            raise RuntimeError("No database connected")

        # Manipulate wkt string for function requirements
        corridor_poly: Polygon = self.data.corridor_poly
        poly_wkt = to_wkt(corridor_poly)
        func_poly_str = poly_wkt.replace("POLYGON ((", "")[:-2]

        query_str = (
            "SELECT "
            "utm_point, "
            "depth, "
            "layer_order, "
            "initial_depth, "
            "sediment_type "
            "FROM "
            "filter.sp_select_cable_corridor_bathymetry_by_polygon"
            "('{}')"
        ).format(func_poly_str)

        result = self._db.server_execute_query(query_str)
        raw_strata = bathy_records_to_strata(result)

        self.data.corridor = raw_strata


class TidalEnergyInterface(QueryInterface):
    """Interface to filter the database for a selected site and technology."""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Tidal Energy Time Series Filtering Interface"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["site.lease_boundary"]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = ["farm.tidal_series"]

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example:

              optional = ["My:first:variable",
                         ]
        """

        option_list = None

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "tidal_series": "farm.tidal_series",
            "lease_poly": "site.lease_boundary",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """
        if self._db is None:
            raise RuntimeError("No database connected")

        # Manipulate wkt string for function requirements
        lease_poly: Polygon = self.data.lease_poly
        poly_wkt = to_wkt(lease_poly)
        func_poly_str = poly_wkt.replace("POLYGON ((", "")[:-2]

        query_str = (
            "SELECT "
            "utm_point, "
            "measure_date, "
            "measure_time, "
            "u, "
            "v, "
            "turbulence_intensity, "
            "ssh "
            "FROM "
            "filter.sp_select_tidal_energy_time_series_by_polygon"
            "('{}')"
        ).format(func_poly_str)

        result = self._db.server_execute_query(query_str)
        raw_strata = tidal_series_records_to_xset(result)

        self.data.tidal_series = raw_strata



# -*- coding: utf-8 -*-

#    Copyright (C) 2016 Mathew Topper, Vincenzo Nava
#    Copyright (C) 2017-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from numbers import Number
from typing import Optional, Sequence

from .base import Strategy
from .basic import BasicStrategy

# Set up logging
module_logger = logging.getLogger(__name__)


class UnitSensitivity(Strategy):
    """A sensitivity study on a single unit variables over a given range of
    values, adjusted before execution of a chosen module."""

    def __init__(self):
        super(UnitSensitivity, self).__init__()

        # Borrow the Basic strategy
        self._basic = BasicStrategy()

    @property
    def version(self) -> int:
        """Version identifier for backwards compatibility when deserializing"""
        return 1

    @classmethod
    def get_name(cls):
        return "Unit Sensitivity"

    def configure(
        self,
        module_name: str,
        variable_name: str,
        variable_values: Sequence[Number],
        skip_errors=True,
    ):
        config_dict = {
            "module_name": module_name,
            "var_name": variable_name,
            "var_values": variable_values,
            "skip_errors": skip_errors,
        }

        self.set_config(config_dict)

    def get_variables(self):
        if self._config is None:
            return []
        return [self._config["var_name"]]

    def execute(self, core, project):
        # Test for Nones
        if (
            self._config is None
            or self._config["module_name"] is None
            or self._config["var_name"] is None
            or self._config["var_values"] is None
        ):
            errStr = (
                "Some configuration values are None. Have you called "
                "the configure method?"
            )
            raise ValueError(errStr)

        module_name = self._config["module_name"]
        variable_name = self._config["var_name"]
        variable_values = self._config["var_values"]

        # Pick up the branch
        if module_name not in self._module_menu.get_available(core, project):
            errStr = "Module {} does not exist".format(module_name)
            raise ValueError(errStr)

        if module_name not in self._module_menu.get_active(core, project):
            errStr = "Module {} has not been activated".format(module_name)
            raise ValueError(errStr)

        mod_branch = self._tree.get_branch(core, project, module_name)

        # Check for existance of the variable
        module_inputs = mod_branch.get_input_status(core, project)

        msgStr = (
            f"Variable {variable_name} is not an input to module {module_name}."
        )
        if module_inputs is None:
            raise ValueError(msgStr)
        else:
            assert isinstance(module_inputs, dict)
            if variable_name not in module_inputs.keys():
                raise ValueError(msgStr)

        unit_var = mod_branch.get_input_variable(core, project, variable_name)

        assert unit_var is not None
        unit_meta = unit_var.get_metadata(core)

        # Check the project is active and record the simulation number
        sim_index = project.get_active_index()

        if sim_index is None:
            errStr = "Project has not been activated."
            raise RuntimeError(errStr)

        sim_titles = []

        # Iterate through the values up to last entry
        for unit_value in variable_values[:-1]:
            # Set the variable
            new_title = self._get_title_str(unit_meta, unit_value)

            # Deal with identical titles
            if new_title in sim_titles:
                n_reps = sum((x.count(new_title) for x in sim_titles))
                new_title = "{} [repeat {}]".format(new_title, n_reps)

            sim_titles.append(new_title)

            project.set_simulation_title(new_title)

            unit_var.set_raw_interface(core, unit_value)
            unit_var.read(core, project)

            success_flag = self._safe_exe(core, project, new_title)

            # Move to the required branch and create a new simulation clone
            if success_flag:
                self.add_simulation_title(new_title)
                core.clone_simulation(project)
                sim_index = project.get_active_index()

            mod_branch.reset(core, project)

        # Set the last variable
        new_title = self._get_title_str(unit_meta, variable_values[-1])

        # Deal with identical titles
        if new_title in sim_titles:
            n_reps = sum((x.count(new_title) for x in sim_titles))
            new_title = "{} [repeat {}]".format(new_title, n_reps)

        project.set_simulation_title(new_title)

        unit_var.set_raw_interface(core, variable_values[-1])
        unit_var.read(core, project)

        # Run the simulation
        success_flag = self._safe_exe(core, project, new_title)

        if success_flag:
            self.add_simulation_title(new_title)

    def _safe_exe(self, core, project, sim_title):
        msg = 'Executing simulation "{}"'.format(sim_title)
        module_logger.info(msg)

        success_flag = True

        assert self._config is not None
        if self._config["skip_errors"]:
            try:
                # Run the simulation
                self._basic.execute(core, project)

            except (KeyboardInterrupt, SystemExit) as e:
                raise e

            except BaseException as e:
                msg = ("Passing exception '{}' for simulation " "{}").format(
                    type(e).__name__, sim_title
                )
                module_logger.exception(msg)

                success_flag = False

        else:
            # Run the simulation
            self._basic.execute(core, project)

        return success_flag

    def _get_title_str(self, meta, value):
        title_str = "{} = {}".format(meta.title, value)

        if meta.units is not None:
            title_str = "{} ({})".format(title_str, meta.units[0])

        return title_str

    @staticmethod
    def dump_config(config: Optional[dict]):
        return config

    @staticmethod
    def load_config(serial_config: Optional[dict], version: int):
        if version != 1:
            raise RuntimeError("Data version not recognised")

        return serial_config

    @staticmethod
    def dump_sim_details(sim_details):
        return None

    @staticmethod
    def load_sim_details(serial_sim_details):
        return None



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2025 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

import pandas as pd
from dtocean_app.utils.display import is_high_dpi
from dtocean_app.widgets.extendedcombobox import ExtendedComboBox
from mdo_engine.utilities.misc import OrderedSet
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt

from dtocean_core.pipeline import Tree
from dtocean_plugins.strategies.multi import MultiSensitivity
from dtocean_plugins.strategy_guis.base import GUIStrategy, StrategyWidget

if is_high_dpi() or TYPE_CHECKING:
    from dtocean_app.designer.high.multisensitivity import (
        Ui_MultiSensitivityWidget,
    )
else:
    from dtocean_app.designer.low.multisensitivity import (
        Ui_MultiSensitivityWidget,
    )


class GUIMultiSensitivity(GUIStrategy, MultiSensitivity):
    """A multi-variable sensitivity study over a given range of
    values, adjusted before execution of a chosen module."""

    def __init__(self):
        MultiSensitivity.__init__(self)
        GUIStrategy.__init__(self)

    def allow_run(self, core, project):
        if len(project) > 1:
            return False
        return True

    def get_weight(self):
        """A method for getting the order of priority of the strategy.

        Returns:
          int
        """

        return 3

    def get_widget(self, parent, shell):
        widget = MultiSensitivityWidget(parent, shell)
        return widget


class MultiSensitivityWidget(
    QtWidgets.QWidget,
    Ui_MultiSensitivityWidget,
    StrategyWidget,
):
    config_set = QtCore.Signal()
    config_null = QtCore.Signal()
    reset = QtCore.Signal()

    def __init__(self, parent, shell):
        QtWidgets.QWidget.__init__(self, parent)
        Ui_MultiSensitivityWidget.__init__(self)
        StrategyWidget.__init__(self)

        self._var_ids = None
        self._mod_names

        self._sim_info_str = (
            "The number of simulations which will be run " "is: {}"
        )

        self._init_ui()
        self._set_interfaces(shell)

    def _init_ui(self):
        self.setupUi(self)

        # Custom boxes
        self.modBox = ExtendedComboBox(self)
        self.modBox.setObjectName("modBox")
        self.gridLayout.addWidget(self.modBox, 0, 1, 1, 1)

        self.varBox = ExtendedComboBox(self)
        self.varBox.setObjectName("varBox")
        self.gridLayout.addWidget(self.varBox, 1, 1, 1, 1)

        # Disble
        self.lineEdit.setDisabled(True)
        self.addButton.setDisabled(True)
        self.removeButton.setDisabled(True)
        self.infoLabel.clear()

        # Set up table model
        tablemodel = SimTableModel(parent=self)
        self.tableView.setModel(tablemodel)

        # Signals
        self.modBox.currentIndexChanged.connect(self._set_variables)
        self.varBox.currentIndexChanged.connect(self._line_edit_ui_switch)
        self.lineEdit.textChanged.connect(self._text_changed_ui_switch)
        self.tableView.clicked.connect(self._table_clicked_ui_switch)
        self.addButton.clicked.connect(self._add_row)
        self.removeButton.clicked.connect(self._remove_row)

    def _get_var_id(self, var_name):
        if self._var_ids is None:
            return var_name
        return self._var_ids[var_name]

    def _set_interfaces(self, shell, include_str=True):
        self._shell = shell

        self.modBox.clear()

        active_modules = shell.module_menu.get_active(shell.core, shell.project)

        self._mod_names = active_modules

        self.modBox.addItems(active_modules)
        self.modBox.setCurrentIndex(-1)

    @QtCore.Slot(int)
    def _set_variables(self, box_number):
        self.varBox.clear()

        if box_number < 0:
            return

        interface_name = self._mod_names[box_number]

        tree = Tree()

        if self._var_ids is None:
            var_id_dict = {}
        else:
            var_id_dict = self._var_ids

        branch = tree.get_branch(
            self._shell.core, self._shell.project, interface_name
        )

        var_inputs = branch.get_inputs(self._shell.core, self._shell.project)

        unique_vars = OrderedSet(var_inputs)

        var_names = []

        for var_id in unique_vars:
            var_meta = self._shell.core.get_metadata(var_id)

            if "SimpleData" in var_meta.structure:
                if var_meta.types is None:
                    errStr = (
                        "Variable {} with SimpleData structure "
                        "requires types meta data to be "
                        "set"
                    ).format(var_id)
                    raise ValueError(errStr)

                if (
                    "int" in var_meta.types
                    or "float" in var_meta.types
                    or "str" in var_meta.types
                ):
                    title = var_meta.title

                    if var_meta.units is not None:
                        title = "{} ({})".format(title, var_meta.units[0])

                    var_names.append(title)

                if title not in var_id_dict:
                    var_id_dict[title] = var_id

        self._var_ids = var_id_dict

        self.varBox.addItems(var_names)
        self.varBox.setCurrentIndex(-1)

    @QtCore.Slot(int)
    def _line_edit_ui_switch(self, box_number):
        self.lineEdit.clear()

        if box_number != -1:
            self.lineEdit.setEnabled(True)
        else:
            self.lineEdit.setDisabled(True)

    @QtCore.Slot()
    def _text_changed_ui_switch(self):
        sane_str = str(self.lineEdit.text())

        if sane_str:
            self.addButton.setEnabled(True)
        else:
            self.addButton.setDisabled(True)

    @QtCore.Slot()
    def _table_clicked_ui_switch(self):
        selected_rows = self._get_selected_rows()

        if not selected_rows:
            self.removeButton.setDisabled(True)
        else:
            self.removeButton.setEnabled(True)

    @QtCore.Slot()
    def _add_row(self):
        model = self.tableView.model()
        assert isinstance(model, SimTableModel)

        last_row = model.rowCount()
        conf_record = self._get_conf_record()
        model.insertRows(last_row, [conf_record])

        self._emit_config_signal()
        self._clear_selections()

    @QtCore.Slot()
    def _remove_row(self):
        model = self.tableView.model()
        assert isinstance(model, SimTableModel)

        selected_row = self._get_selected_rows()[0]
        model.removeRows(selected_row.row())

        self._table_clicked_ui_switch()
        self._emit_config_signal()

    def _emit_config_signal(self):
        model = self.tableView.model()
        assert isinstance(model, SimTableModel)

        df = model.array_df

        if not df.empty:
            self.config_set.emit()
        else:
            self.config_null.emit()

    def _get_conf_record(self):
        mod_name = str(self.modBox.currentText())
        var_name = str(self.varBox.currentText())
        var_values = str(self.lineEdit.text())

        conf_record = (mod_name, var_name, var_values)

        return conf_record

    def _get_selected_rows(self):
        indexes = self.tableView.selectionModel().selectedRows()

        return indexes

    def _clear_selections(self):
        self.modBox.setCurrentIndex(-1)
        self.varBox.clear()
        self.lineEdit.clear()

    def _get_var_title(self, var_id):
        var_meta = self._shell.core.get_metadata(var_id)
        var_title = var_meta.title

        if var_meta.units is not None:
            var_title = "{} ({})".format(var_title, var_meta.units[0])

        return var_title

    def _list2string(self, typed_list):
        sane_var_values = [str(x) for x in typed_list]
        var_values_str = ", ".join(sane_var_values)

        return var_values_str

    def get_configuration(self):
        model = self.tableView.model()
        assert isinstance(model, SimTableModel)

        df = model.array_df.copy()
        df["Variable"] = df["Variable"].apply(self._get_var_id)
        df["Values"] = df["Values"].apply(lambda x: self.string2types(x))

        subsp_ratio = self.subsetSpinBox.value() / 100.0

        conf_dict = {"inputs_df": df, "subspacing_ratio": subsp_ratio}

        nsims = MultiSensitivity.count_selections(df, subsp_ratio)
        info_str = self._sim_info_str.format(nsims)

        self.infoLabel.setText(info_str)

        return conf_dict

    def set_configuration(self, config_dict=None):
        if config_dict is None:
            return

        model = self.tableView.model()
        assert isinstance(model, SimTableModel)

        df = config_dict["inputs_df"]
        subsp_ratio = config_dict["subsp_ratio"]
        nsims = MultiSensitivity.count_selections(df, subsp_ratio)

        df["Variable"] = df["Variable"].apply(lambda x: self._get_var_title(x))
        df["Values"] = df["Values"].apply(lambda x: self._list2string(x))

        model.array_df = df
        self.subsetSpinBox.setValue(subsp_ratio * 100.0)

        info_str = self._sim_info_str.format(nsims)
        self.infoLabel.setText(info_str)

        self._emit_config_signal()


class SimTableModel(QtCore.QAbstractTableModel):
    header_labels = ["Module", "Variable", "Values"]

    def __init__(self, init_data=None, parent=None, *args):
        QtCore.QAbstractTableModel.__init__(self, parent, *args)
        self.array_df = self._init_df(init_data)

    def _init_df(self, init_records=None):
        new_df = pd.DataFrame(columns=self.header_labels)

        if init_records is None:
            return new_df

        new_df = self._add_records(new_df, init_records)

        return new_df

    def headerData(
        self,
        section,
        orientation: Qt.Orientation,
        role=Qt.ItemDataRole.DisplayRole,
    ):
        if (
            role == Qt.ItemDataRole.DisplayRole
            and orientation == Qt.Orientation.Horizontal
        ):
            return self.header_labels[section]

        return QtCore.QAbstractTableModel.headerData(
            self, section, orientation, role
        )

    def insertRows(self, position, records, rows=1, index=QtCore.QModelIndex()):
        self.beginInsertRows(
            QtCore.QModelIndex(), position, position + rows - 1
        )

        self.array_df = self._add_records(self.array_df, records)

        self.endInsertRows()

        return True

    def removeRows(self, position, rows=1, index=QtCore.QModelIndex()):
        self.beginRemoveRows(
            QtCore.QModelIndex(), position, position + rows - 1
        )

        drop_range = range(position, position + rows)
        drop_indexes = self.array_df.index[drop_range]
        self.array_df = self.array_df.drop(drop_indexes)

        self.endRemoveRows()

        return True

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.array_df)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 3

    def data(self, index, role):
        if not index.isValid() or role != Qt.ItemDataRole.DisplayRole:
            return None

        value = self.array_df.iloc[index.row(), index.column()]

        return value

    def _add_records(self, df: pd.DataFrame, records):
        for record in records:
            s2 = pd.Series(record, index=self.header_labels)
            df = pd.concat([df, s2], ignore_index=True)

        return df



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Wed Apr 06 15:59:04 2016

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import math
from typing import Optional

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import numpy as np
import shapely.geometry as sgeom
from cartopy.mpl.geoaxes import GeoAxes
from matplotlib import ticker
from pyproj import Transformer
from shapely import Polygon, transform
from shapely.plotting import patch_from_polygon

from .base import PlotInterface

BLUE = "#6699cc"
GREEN = "#32CD32"
RED = "#B20000"


class SiteBoundaryPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Site Boundary"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["site.selected_name", "hidden.site_boundaries"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "name": "site.selected_name",
            "boundaries": "hidden.site_boundaries",
        }

        return id_map

    def connect(self):
        def down(number, base=5):
            return base * math.ceil(number / base)

        def up(number, base=5):
            return base * (math.ceil(number / base) + 1)

        site_poly = self.data.boundaries[self.data.name]

        centroid = np.array(site_poly.centroid.coords[0])
        left = centroid[0] - 5.0
        right = centroid[0] + 5.0
        bottom = centroid[1] - 5.0
        top = centroid[1] + 5.0

        proj = ccrs.LambertConformal(
            central_longitude=centroid[0],
            central_latitude=centroid[1],
        )

        # Base plot
        ax = plt.axes(projection=proj, facecolor=cfeature.COLORS["water"])
        assert isinstance(ax, GeoAxes)
        ax.set_extent([left, right, bottom, top], ccrs.PlateCarree())

        feat = cfeature.LAND.with_scale("50m")
        geoms = intersecting_geometries(feat, [left, right, bottom, top])
        ax.add_geometries(
            geoms,
            crs=ccrs.PlateCarree(),
            facecolor=cfeature.COLORS["land"],
        )

        feat = cfeature.COASTLINE.with_scale("50m")
        geoms = intersecting_geometries(feat, [left, right, bottom, top])
        ax.add_geometries(
            geoms,
            crs=ccrs.PlateCarree(),
            edgecolor="black",
            facecolor="never",
        )

        feat = cfeature.BORDERS.with_scale("50m")
        geoms = intersecting_geometries(feat, [left, right, bottom, top])
        ax.add_geometries(
            geoms,
            crs=ccrs.PlateCarree(),
            edgecolor="black",
            facecolor="never",
        )

        # Gridlines
        gl = ax.gridlines(
            crs=ccrs.PlateCarree(),
            draw_labels=True,
            x_inline=False,
            y_inline=False,
            linewidth=0.33,
            color="k",
            alpha=0.5,
        )
        gl.right_labels = gl.top_labels = False
        gl.ylocator = ticker.FixedLocator(np.arange(down(bottom), up(top), 5))
        gl.xlocator = ticker.FixedLocator(np.arange(down(left), up(right), 5))

        # Annotation
        ax.plot(
            centroid[0],
            centroid[1],
            "s",
            mew=3,
            ms=20,
            fillstyle="none",
            markeredgecolor="orange",
            transform=ccrs.PlateCarree(),
        )
        ax.annotate(
            self.data.name,
            xy=(centroid[0], centroid[1] + 1),
            horizontalalignment="center",
            color="black",
            backgroundcolor="white",
            size="large",
            weight="roman",
            transform=ccrs.PlateCarree(),
        )

        self.fig_handle = plt.gcf()


def intersecting_geometries(feat, extent):
    geometries = feat.geometries()
    if extent is not None and not np.isnan(extent[0]):
        extent_geom = sgeom.box(extent[0], extent[2], extent[1], extent[3])
        return (
            geom
            for geom in geometries
            if geom is not None and extent_geom.intersects(geom)
        )
    else:
        return geometries


class AllBoundaryPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "All Boundaries"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "hidden.site_boundary",
            "site.lease_boundary",
            "site.corridor_boundary",
            "corridor.landing_point",
            "site.projection",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = [
            "site.lease_boundary",
            "site.corridor_boundary",
            "corridor.landing_point",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "site_poly": "hidden.site_boundary",
            "lease_poly": "site.lease_boundary",
            "corridor_poly": "site.corridor_boundary",
            "landing_point": "corridor.landing_point",
            "projection": "site.projection",
        }

        return id_map

    def connect(self):
        self.fig_handle = boundaries_plot(
            site_poly=self.data.site_poly,
            projection=self.data.projection,
            lease_poly=self.data.lease_poly,
            corridor_poly=self.data.corridor_poly,
            landing_point=self.data.landing_point,
        )


class DesignBoundaryPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Design Boundaries"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.lease_boundary",
            "site.corridor_boundary",
            "farm.nogo_areas",
            "corridor.nogo_areas",
            "corridor.landing_point",
            "project.lease_area_entry_point",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = [
            "site.lease_boundary",
            "site.corridor_boundary",
            "farm.nogo_areas",
            "corridor.nogo_areas",
            "corridor.landing_point",
            "project.lease_area_entry_point",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "lease_poly": "site.lease_boundary",
            "corridor_poly": "site.corridor_boundary",
            "landing_point": "corridor.landing_point",
            "lease_entry_point": "project.lease_area_entry_point",
            "nogo_areas": "farm.nogo_areas",
            "corridor_nogo_areas": "corridor.nogo_areas",
        }

        return id_map

    def connect(self):
        self.fig_handle = boundaries_plot(
            lease_poly=self.data.lease_poly,
            corridor_poly=self.data.corridor_poly,
            nogo_polys=self.data.nogo_areas,
            corridor_nogo_polys=self.data.corridor_nogo_areas,
            landing_point=self.data.landing_point,
            lease_entry_point=self.data.lease_entry_point,
        )


def boundaries_plot(
    site_poly: Optional[Polygon] = None,
    projection=None,
    lease_poly=None,
    corridor_poly: Optional[Polygon] = None,
    nogo_polys: Optional[dict[str, Polygon]] = None,
    corridor_nogo_polys=None,
    landing_point=None,
    lease_entry_point=None,
):
    if site_poly is None and lease_poly is None and corridor_poly is None:
        return None

    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1, aspect="equal")

    # Convert the site polygon to local coordinate system
    if site_poly is not None:
        transformer = Transformer.from_crs(
            "EPSG:4326", projection, always_xy=True
        )

        def transform_ndarray(x: np.ndarray):
            transformed = transformer.transform(x[:, 0], x[:, 1])
            return np.array(transformed).T

        local_site_poly = transform(
            site_poly,
            transform_ndarray,
            include_z=False,
        )

        patch = patch_from_polygon(
            local_site_poly,
            fc=RED,
            ec=RED,
            fill=False,
            linewidth=1,
        )
        ax1.add_patch(patch)

    if lease_poly is not None:
        patch = patch_from_polygon(
            lease_poly,
            fc=BLUE,
            ec=BLUE,
            fill=False,
            linewidth=2,
        )
        ax1.add_patch(patch)

        maxy = lease_poly.bounds[3] + 50.0
        centroid = np.array(lease_poly.centroid.coords[0])

        ax1.annotate(
            "Lease Area",
            xy=(centroid[0], maxy),
            horizontalalignment="center",
            verticalalignment="bottom",
            weight="bold",
            size="large",
        )

    if corridor_poly is not None:
        patch = patch_from_polygon(
            corridor_poly,
            fc=GREEN,
            ec=GREEN,
            fill=False,
            linewidth=2,
        )
        ax1.add_patch(patch)

        miny = corridor_poly.bounds[1] - 50.0
        centroid = np.array(corridor_poly.centroid.coords[0])

        ax1.annotate(
            "Cable Corridor",
            xy=(centroid[0], miny),
            horizontalalignment="center",
            verticalalignment="top",
            weight="bold",
            size="large",
        )

    if nogo_polys is not None:
        for key, polygon in nogo_polys.items():
            patch = patch_from_polygon(
                polygon,
                fc=RED,
                ec=RED,
                fill=True,
                alpha=0.3,
                linewidth=2,
            )
            ax1.add_patch(patch)

            centroid = np.array(polygon.centroid.coords[0])
            ax1.annotate(
                key,
                xy=tuple(centroid[:2]),
                xytext=(0, 0),
                xycoords="data",
                textcoords="offset pixels",
                horizontalalignment="center",
                weight="bold",
                size="large",
            )

    if corridor_nogo_polys is not None:
        for key, polygon in corridor_nogo_polys.items():
            patch = patch_from_polygon(
                polygon,
                fc=RED,
                ec=RED,
                fill=True,
                alpha=0.3,
                linewidth=2,
            )
            ax1.add_patch(patch)

            centroid = np.array(polygon.centroid.coords[0])
            ax1.annotate(
                key,
                xy=tuple(centroid[:2]),
                xytext=(0, 0),
                xycoords="data",
                textcoords="offset pixels",
                horizontalalignment="center",
                weight="bold",
                size="large",
            )

    if landing_point is not None:
        xy = list(landing_point.coords)[0]
        ax1.plot(xy[0], xy[1], "or")

        label_xy = (xy[0] + 50, xy[1])

        ax1.annotate(
            "Export Cable Landing",
            xy=label_xy,
            horizontalalignment="left",
            verticalalignment="center",
            weight="bold",
            size="large",
        )

    if lease_entry_point is not None:
        xy = list(lease_entry_point.coords)[0]
        ax1.plot(xy[0], xy[1], "sb")

        label_xy = (xy[0] + 50, xy[1])

        ax1.annotate(
            "Lease Area Entry Point",
            xy=label_xy,
            horizontalalignment="left",
            verticalalignment="center",
            weight="bold",
            size="large",
        )

    ax1.margins(0.1, 0.1)
    ax1.autoscale_view()

    xlabel = "UTM x (m)"
    ylabel = "UTM y (m)"

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.ticklabel_format(useOffset=False)

    return fig



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Tue Feb 23 15:38:18 2016

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import logging
import os
import pickle
import shutil
import tarfile
import tempfile
import time
import zipfile
from collections.abc import Callable
from pathlib import Path
from typing import Protocol, Union

StrOrPath = Union[str, Path]

# Set up logging
module_logger = logging.getLogger(__name__)


def pickle_test_data(file_path, test_data_dict):
    dir_path = os.path.dirname(file_path)
    file_name = os.path.basename(file_path)
    file_root = os.path.splitext(file_name)[0]
    dst_name = "{}.pkl".format(file_root)
    dst_path = os.path.join(dir_path, dst_name)

    with open(dst_path, "wb") as dataf:
        # Pickle dictionary using protocol 0.
        pickle.dump(test_data_dict, dataf)

    return dst_path


def package_dir(src_dir_path, dst_path: StrOrPath, archive=False):
    dst_path = Path(dst_path)

    if not os.path.splitext(dst_path)[1]:
        archive = False

    if not archive:
        for root, dirs, files in os.walk(src_dir_path, topdown=True):
            for name in dirs:
                new_path = dst_path / name
                if os.path.exists(new_path):
                    shutil.rmtree(new_path)
                os.makedirs(new_path)

            for name in files:
                tmp_path = os.path.join(root, name)
                short_path = tmp_path.replace(src_dir_path, "")
                new_path = os.path.join(dst_path, short_path[1:])
                shutil.move(tmp_path, new_path)

        shutil.rmtree(src_dir_path)

        return

    tgz_dir_path = tempfile.mkdtemp()

    project_file_name = os.path.split(dst_path)[1]
    tgz_file_name = "{}.tar.gz".format(project_file_name)
    tgz_file_path = os.path.join(tgz_dir_path, tgz_file_name)

    with tarfile.open(tgz_file_path, "w:gz") as tar:
        for root, dirs, files in os.walk(src_dir_path):
            for name in files:
                file_path = os.path.join(root, name)
                short_path = file_path.replace(src_dir_path, "")
                tar.add(file_path, arcname=short_path)

    shutil.copyfile(tgz_file_path, dst_path)

    rmtree_retry(src_dir_path, fail_silent=True)
    rmtree_retry(tgz_dir_path, fail_silent=True)


def unpack_archive(src_path, dst_path):
    # Determine if archive is new tar style or legacy zip
    if tarfile.is_tarfile(src_path):
        tar = tarfile.open(src_path, "r:gz")
        tar.extractall(dst_path, filter="data")
        tar.close()
    else:
        zf = zipfile.ZipFile(src_path, "r")
        zf.extractall(dst_path)
        zf.close()


def onerror(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``

    Copyright Michael Foord 2004
    Released subject to the BSD License
    Please see http://www.voidspace.org.uk/python/license.shtml
    """
    import stat

    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


class OSRetry(Protocol):
    def __call__(
        self,
        src_path: str,
        max_attempts: int = 60,
        sleep_seconds: int = 1,
        fail_silent: bool = False,
    ): ...


def os_retry(func: Callable[[str], None]) -> OSRetry:
    def wrapper(src_path, max_attempts=60, sleep_seconds=1, fail_silent=False):
        file_locked = True
        n_attempts = 0

        while file_locked:
            n_attempts += 1

            if n_attempts > max_attempts:
                if fail_silent:
                    return

                total_seconds = max_attempts * sleep_seconds

                err_str = (
                    "shutil.rmtree failed for over {} seconds: " "{}"
                ).format(total_seconds, src_path)
                raise OSError(err_str)

            try:
                func(src_path)
                file_locked = False
            except OSError:
                time.sleep(sleep_seconds)

    return wrapper


@os_retry
def rmtree_retry(src_path):
    shutil.rmtree(src_path)


@os_retry
def remove_retry(src_path):
    os.remove(src_path)


def init_dir(dir_name, clean_existing=False):
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
        return

    if not clean_existing:
        err_msg = (
            "Directory {} already exists. Set clean_existing "
            "argument to True to delete the contents of the "
            "directory"
        ).format(dir_name)
        raise IOError(err_msg)

    for entry_name in os.listdir(dir_name):
        entry_path = os.path.join(dir_name, entry_name)

        if os.path.isfile(entry_path):
            remove_retry(entry_path)
        elif os.path.isdir(entry_path):
            rmtree_retry(entry_path)
        else:
            err_str = ("Path {} has unhandled filesystem " "type").format(
                entry_path
            )
            raise RuntimeError(err_str)



# -*- coding: utf-8 -*-

#    Copyright (c) 2014 Inria
#    Copyright (C) 2021-2026 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#  Author: Nikolaus Hansen, 2008-
#  Author: Petr Baudis, 2014
#  Author: Youhei Akimoto, 2016-
#  Author: Mathew Topper 2021-

# pylint: disable=line-too-long

from collections import deque

import numpy as np
from cma.utilities.math import Mh


class MovingAverage:
    def __init__(self, window_length, init=None):
        if init is None:
            init = []

        self._length = window_length
        self._data = deque(init, window_length)
        self._predict = None

    def add(self, new_data=None):
        if new_data is not None:
            self._data.append(new_data)

        if len(self._data) < self._length:
            result = None
        else:
            result = sum(self._data) / float(self._length)

        self._predict = result

    def __call__(self):
        return self._predict


class NoiseHandler:
    """Noise handling according to [Hansen et al 2009, A Method for
    Handling Uncertainty in Evolutionary Optimization...]
    The interface of this class is yet versatile and subject to changes.
    The noise handling follows closely [Hansen et al 2009] in the
    measurement part, but the implemented treatment is slightly
    different: for ``noiseS > 0``, ``evaluations`` (time) and sigma are
    increased by ``alpha``. For ``noiseS < 0``, ``evaluations`` (time)
    is decreased by ``alpha**(1/4)``.
    The (second) parameter ``evaluations`` defines the maximal number
    of evaluations for a single fitness computation. If it is a list,
    the smallest element defines the minimal number and if the list has
    three elements, the median value is the start value for
    ``evaluations``.
    `NoiseHandler` serves to control the noise via steps-size
    increase and number of re-evaluations, for example via `fmin` or
    with `ask_and_eval`.
    Examples
    --------
    Minimal example together with `fmin` on a non-noisy function:
    >>> import cma
    >>> res = cma.fmin(cma.ff.elli, 7 * [1], 1, noise_handler=cma.NoiseHandler(7))  #doctest: +ELLIPSIS
    (4_w,9)-aCMA-ES (mu_w=2.8,...
    >>> assert res[1] < 1e-8
    >>> res = cma.fmin(cma.ff.elli, 6 * [1], 1, {'AdaptSigma':cma.sigma_adaptation.CMAAdaptSigmaTPA},
    ...          noise_handler=cma.NoiseHandler(6))  #doctest: +ELLIPSIS
    (4_w,...
    >>> assert res[1] < 1e-8
    in dimension 7 (which needs to be given tice). More verbose example
    in the optimization loop with a noisy function defined in ``func``:
    >>> import cma, numpy as np
    >>> func = lambda x: cma.ff.sphere(x) * (1 + 4 * np.random.randn() / len(x))  # cma.ff.noisysphere
    >>> es = cma.CMAEvolutionStrategy(np.ones(10), 1)  #doctest: +ELLIPSIS
    (5_w,10)-aCMA-ES (mu_w=3.2,...
    >>> nh = cma.NoiseHandler(es.N, maxevals=[1, 1, 30])
    >>> while not es.stop():
    ...     X, fit_vals = es.ask_and_eval(func, evaluations=nh.evaluations)
    ...     es.tell(X, fit_vals)  # prepare for next iteration
    ...     es.sigma *= nh(X, fit_vals, func, es.ask)  # see method __call__
    ...     es.countevals += nh.evaluations_just_done  # this is a hack, not important though
    ...     es.logger.add(more_data = [nh.evaluations, nh.noiseS])  # add a data point
    ...     es.disp()
    ...     # nh.maxevals = ...  it might be useful to start with smaller values and then increase
    ...                # doctest: +ELLIPSIS
    Iterat...
    >>> print(es.stop())
    ...                # doctest: +ELLIPSIS
    {...
    >>> print(es.result[-2])  # take mean value, the best solution is totally off
    ...                # doctest: +ELLIPSIS
    [...
    >>> assert sum(es.result[-2]**2) < 1e-9
    >>> print(X[np.argmin(fit_vals)])  # not bad, but probably worse than the mean
    ...                # doctest: +ELLIPSIS
    [...
    >>> # es.logger.plot()
    The command ``logger.plot()`` will plot the logged data.
    The noise options of fmin` control a `NoiseHandler` instance
    similar to this example. The command ``cma.CMAOptions('noise')``
    lists in effect the parameters of `__init__` apart from
    ``aggregate``.
    Details
    -------
    The parameters reevals, theta, c_s, and alpha_t are set differently
    than in the original publication, see method `__init__`. For a
    very small population size, say popsize <= 5, the measurement
    technique based on rank changes is likely to fail.
    Missing Features
    ----------------
    In case no noise is found, ``self.lam_reeval`` should be adaptive
    and get at least as low as 1 (however the possible savings from this
    are rather limited). Another option might be to decide during the
    first call by a quantitative analysis of fitness values whether
    ``lam_reeval`` is set to zero. More generally, an automatic noise
    mode detection might also set the covariance matrix learning rates
    to smaller values.
    :See also: `fmin`, `CMAEvolutionStrategy.ask_and_eval`
    """

    # TODO: for const additive noise a better version might be with alphasigma
    # also used for sigma-increment, while all other variance changing sources
    # are removed (because they are intrinsically biased). Then using kappa to
    # get convergence (with unit sphere samples): noiseS=0 leads to a certain
    # kappa increasing rate?
    def __init__(self, N, maxevals=None, reevals=None, epsilon=1e-7):
        """Parameters are:
        ``N``
            dimension, (only) necessary to adjust the internal
            "alpha"-parameters
        ``maxevals``
            maximal value for ``self.evaluations``, where
            ``self.evaluations`` function calls are aggregated for
            noise treatment. With ``maxevals == 0`` the noise
            handler is (temporarily) "switched off". If `maxevals`
            is a list, min value and (for >2 elements) median are
            used to define minimal and initial value of
            ``self.evaluations``. Choosing ``maxevals > 1`` is only
            reasonable, if also the original ``fit`` values (that
            are passed to `__call__`) are computed by aggregation of
            ``self.evaluations`` values (otherwise the values are
            not comparable), as it is done within `fmin`.
        ``reevals``
            number of solutions to be reevaluated for noise
            measurement, can be a float, by default set to ``2 +
            popsize/20``, where ``popsize = len(fit)`` in
            ``__call__``. zero switches noise handling off.
        ``epsilon``
            multiplier for perturbation of the reevaluated solutions
        :See also: `fmin`, `CMAOptions`, `CMAEvolutionStrategy.ask_and_eval`
        """

        if maxevals is None:
            maxevals = [1, 1, 1]

        self.lam_reeval = reevals  # 2 + popsize/20, see method indices(), originally 2 + popsize/10
        self.epsilon = epsilon
        ## meta_parameters.noise_theta == 0.5
        self.theta = 0.5  # 0.5  # originally 0.2
        self.cum = 0.3  # originally 1, 0.3 allows one disagreement of current point with resulting noiseS
        ## meta_parameters.noise_alphasigma == 2.0
        self.alphasigma = 1 + 1.0 / (
            N + 10
        )  # 2, unit sphere sampling: 1 + 1 / (N + 10)
        ## meta_parameters.noise_alphaevals == 2.0
        self.alphaevals = 1 + 2.0 / (N + 10)  # 2, originally 1.5
        ## meta_parameters.noise_alphaevalsdown_exponent == -0.25
        self.alphaevalsdown = self.alphaevals**-0.25  # originally 1/1.5
        # zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz
        if 11 < 3 and maxevals[2] > 1e18:  # for testing purpose
            self.alphaevals = 1.5
            self.alphaevalsdown = self.alphaevals**-0.999  # originally 1/1.5

        self.evaluations = 1
        """number of f-evaluations to get a single measurement by aggregation"""
        self.minevals = 1
        self.maxevals = int(np.max(maxevals))
        if hasattr(maxevals, "__contains__"):  # i.e. can deal with ``in``
            if len(maxevals) > 1:
                self.minevals = min(maxevals)
                self.evaluations = self.minevals
            if len(maxevals) > 2:
                self.evaluations = np.median(maxevals)
        self.n_evals = max(int(self.evaluations), 1)
        self.last_n_evals = None
        ## meta_parameters.noise_aggregate == None
        self.evaluations_just_done = (
            0  # actually conducted evals, only for documentation
        )
        self.noiseS = 0
        self.idx = None
        self.popsize = None
        self.fit = None
        self.fitre = None
        self._X = None
        self._sols = None
        self._sigma_fac = 1.0
        self._noise_predict = MovingAverage(3)

    @property
    def sigma_fac(self):
        return self._sigma_fac

    def get_predicted_noise(self):
        result = self._noise_predict()
        if result is None:
            result = 0.0

        return result

    def prepare(self, X, fit):
        """Prepare noise measurement.
        Parameters
        ----------
        ``X``
            a list/sequence/vector of solutions
        ``fit``
            the respective list of function values
        Details
        -------
        Calls the methods `reeval`, `update_measure` and ``treat` in
        this order. ``self.evaluations`` is adapted within the method
        `treat`.
        """

        self.last_n_evals = self.n_evals
        self.n_evals = int(self.evaluations)
        self.evaluations_just_done = 0
        self._sols = None

        if not self.maxevals or self.lam_reeval == 0:
            return

        self.idx = self._indices(list(fit))
        self.popsize = len(self.idx)

        if self.idx.size == 0:
            return

        self.fit = list(fit)
        self.fitre = list(fit)
        self._X = X

    def ask(self, ask=None):
        """Give solutions for noise measurement.
        Parameters
        ----------
        ``ask``
            a method to generate a new, slightly disturbed solution. The
            argument is (only) mandatory if ``epsilon`` is not zero, see
            `__init__`.
        """

        if self._X is None or self.idx is None:
            raise RuntimeError("Call prepare first")

        if self.epsilon:
            if ask is None:
                raise ValueError("ask must be defined if epsilon is not zero")
            sols = [ask(1, self._X[i], self.epsilon)[0] for i in self.idx]
        else:
            sols = [self._X[i] for i in self.idx]

        # Store a copy of the solutions for checking in tell
        self._sols = sols[:]

        return sols

    def tell(self, sols, function_values):
        """Get function values for noise measurement, set anew attributes ``evaluations``
        (proposed number of evaluations to "treat" noise) and ``evaluations_just_done``
        and return a factor for increasing sigma.
        """

        def get_sol_match(i, test_sol):
            if self._sols is None:
                raise RuntimeError("Call ask first")

            for j, sol in enumerate(self._sols):
                if np.isclose(test_sol, sol).all():
                    return j, function_values[i]

            test_sol_strs = [str(x) for x in test_sol]
            test_sol_str = ", ".join(test_sol_strs)
            err_msg = "Solution {} was not asked for".format(test_sol_str)
            raise RuntimeError(err_msg)

        # Case where prepare has not been called
        if None in [self.fit, self.fitre]:
            return

        # Case where ask has not been called
        if self._sols is None:
            return

        # No return values given (probably due to constraint violation)
        if not sols or not function_values:
            self.fit = None
            self.fitre = None
            self._sigma_fac = 1.0

            return

        # Check the given solutions and sort the function values
        n_sols = len(self._sols)

        if len(sols) != n_sols or len(function_values) != n_sols:
            err_msg = (
                "Exactly {} solutions and function values must be " "provided"
            ).format(n_sols)
            raise ValueError(err_msg)

        f_sorted = np.zeros(n_sols)

        for i, test_sol in enumerate(sols):
            j, f = get_sol_match(i, test_sol)
            f_sorted[j] = f

        real_evals = 0

        assert self.idx is not None
        assert self.fitre is not None

        for j, i in enumerate(self.idx):
            if np.isnan(f_sorted[j]):
                continue
            self.fitre[i] = f_sorted[j]
            real_evals += 1

        self._update_measure()
        self.evaluations_just_done = real_evals
        self._sigma_fac = self._treat()
        self._sols = None

    def _indices(self, fit):
        """return the set of indices to be reevaluated for noise
        measurement.
        Given the first values are the earliest, this is a useful policy
        also with a time changing objective.
        """
        ## meta_parameters.noise_reeval_multiplier == 1.0
        lam_reev = 1.0 * (
            self.lam_reeval if self.lam_reeval else 2 + len(fit) / 20.0
        )
        lam_reev = int(lam_reev) + ((lam_reev % 1) > np.random.rand())
        ## meta_parameters.noise_choose_reeval == 1
        choice = 1
        if choice == 1:
            # take n_first first and reev - n_first best of the remaining
            n_first = lam_reev - lam_reev // 2
            sort_idx = np.argsort(np.asarray(fit)[n_first:]) + n_first
            return np.asarray(
                list(range(0, n_first)) + list(sort_idx[0 : lam_reev - n_first])
            )
        elif choice == 2:
            idx_sorted = np.argsort(np.asarray(fit))
            # take lam_reev equally spaced, starting with best
            linsp = np.linspace(0, len(fit) - len(fit) / lam_reev, lam_reev)
            return idx_sorted[[int(i) for i in linsp]]
        # take the ``lam_reeval`` best from the first ``2 * lam_reeval + 2`` values.
        elif choice == 3:
            return np.argsort(np.asarray(fit)[: 2 * (lam_reev + 1)])[:lam_reev]
        else:
            raise ValueError(
                "unrecognized choice value %d for noise reev" % choice
            )

    def _update_measure(self):
        """updated noise level measure using two fitness lists ``self.fit`` and
        ``self.fitre``, return ``self.noiseS, all_individual_measures``.
        Assumes that ``self.idx`` contains the indices where the fitness
        lists differ.
        """

        if self.fit is None or self.fitre is None or self.idx is None:
            raise RuntimeError("Call prepare first")

        lam = len(self.fit)
        idx = np.argsort(self.fit + self.fitre)
        ranks = np.argsort(idx).reshape((2, lam))
        rankDelta = ranks[0] - ranks[1] - np.sign(ranks[0] - ranks[1])

        # compute rank change limits using both ranks[0] and ranks[1]
        r = np.arange(1, 2 * lam)  # 2 * lam - 2 elements

        limits = []

        for i in self.idx:
            first = Mh.prctile(
                np.abs(r - (ranks[0, i] + 1 - (ranks[0, i] > ranks[1, i]))),
                self.theta * 50,
            )
            second = Mh.prctile(
                np.abs(r - (ranks[1, i] + 1 - (ranks[1, i] > ranks[0, i]))),
                self.theta * 50,
            )
            assert isinstance(first, float)
            assert isinstance(second, float)

            limit = 0.5 * (first + second)
            limits.append(limit)

        # compute measurement
        # max: 1 rankchange in 2*lambda is always fine
        s = np.abs(rankDelta[self.idx]) - Mh.amax(
            limits, 1
        )  # lives roughly in 0..2*lambda
        self.noiseS += self.cum * (np.mean(s) - self.noiseS)
        self._noise_predict.add(self.noiseS)

        self.fit = None
        self.fitre = None

        return self.noiseS, s

    def _treat(self):
        """adapt self.evaluations depending on the current measurement
        value and return ``sigma_fac in (1.0, self.alphasigma)``
        """

        # Use noise prediction
        predicted_noise = self._noise_predict()

        if predicted_noise is None:
            return 1.0

        if predicted_noise > 0:
            self.evaluations = min(
                (self.evaluations * self.alphaevals, self.maxevals)
            )

            if self.evaluations == self.maxevals:
                sigma_fac = 1.0
            else:
                sigma_fac = self.alphasigma

        else:
            self.evaluations = max(
                (self.evaluations * self.alphaevalsdown, self.minevals)
            )
            sigma_fac = 1.0

        return sigma_fac



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import logging

from polite_config.paths import DirectoryMap, ModPath, UserDataPath

# Set up logging
module_logger = logging.getLogger(__name__)


def init_config(logging=False, database=False, overwrite=False):
    """Copy config files to user data directory"""

    if not any([logging, database]):
        return

    objdir = ModPath(__name__, "..", "config")
    datadir = UserDataPath("dtocean_core", "DTOcean", "config")
    dirmap = DirectoryMap(datadir, objdir)

    if logging:
        dirmap.copy_file("logging.yaml", overwrite=overwrite)
    if database:
        dirmap.copy_file("database.yaml", overwrite=overwrite)

    return datadir



import argparse
import datetime
import importlib
from inspect import getmembers, isfunction
from types import FunctionType, ModuleType

import dtocean_plugins.cli as cli
from mdo_engine.utilities.plugins import get_module_names_from_package


def main():
    now = datetime.datetime.now()
    epiStr = "The DTOcean Developers (c) {}.".format(now.year)

    parser = argparse.ArgumentParser(prog="dtocean", epilog=epiStr)
    subparser = parser.add_subparsers(required=True)
    setup_init(subparser)

    for subcommand in get_plugin_function("subcommand", cli):
        subcommand(subparser)

    args = parser.parse_args()
    args.func(args)


def setup_init(subparser):
    def run_inits(args):
        for f in get_plugin_function("init", cli):
            f(args)

    description = "initialize modules (requires internet connection)"
    parser = subparser.add_parser(
        "init",
        description=description,
        help=description,
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="overwrite existing files",
    )

    parser.set_defaults(func=run_inits)


def get_plugin_function(function_name: str, module: ModuleType):
    result: list[FunctionType] = []

    for name in get_module_names_from_package(module):
        mod = importlib.import_module(name)
        members = getmembers_dict(mod, isfunction)
        if function_name in members:
            result.append(members[function_name])

    return result


def getmembers_dict(mod: ModuleType, predicate=None):
    members = getmembers(mod, predicate=predicate)
    return {n: v for n, v in members}



# -*- coding: utf-8 -*-

#    Copyright (C) 2018-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import cast

import cmocean
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle

from .base import PlotInterface


class CombinedBathyPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Combined Bathymetry"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["bathymetry.layers", "corridor.layers"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "deployment_strata": "bathymetry.layers",
            "cable_corridor_strata": "corridor.layers",
        }

        return id_map

    def connect(self):
        deployment_bathy = self.data.deployment_strata["depth"].sel(
            layer="layer 1"
        )
        cable_corridor_bathy = self.data.cable_corridor_strata["depth"].sel(
            layer="layer 1"
        )

        fig = plt.figure()
        fig.add_subplot(1, 1, 1, aspect="equal")

        step = 5
        bm = np.amin(deployment_bathy)
        bm = int(10 * round(float(bm) / 10))
        levels = np.arange(bm, 0.0 + step, step)

        x = deployment_bathy.coords["x"]
        y = deployment_bathy.coords["y"]

        plt.contourf(
            x,
            y,
            deployment_bathy.T,
            levels=levels,
            cmap=cmocean.cm.cmap_d["deep"],
        )

        x = cable_corridor_bathy.coords["x"]
        y = cable_corridor_bathy.coords["y"]

        plt.contourf(
            x,
            y,
            cable_corridor_bathy.T,
            levels=levels,
            cmap=cmocean.cm.cmap_d["deep"],
        )

        clb = plt.colorbar()

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"
        zlabel = "Depth [$m$]"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        clb.set_label(zlabel)

        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")

        plt.title("Combined Bathymetry")
        plt.gcf().tight_layout(rect=(0.05, 0.05, 0.95, 0.95))

        self.fig_handle = plt.gcf()


class CombinedSedimentPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Combined Sediment (First Layer)"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["bathymetry.layers", "corridor.layers"]

        return input_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "deployment_strata": "bathymetry.layers",
            "cable_corridor_strata": "corridor.layers",
        }

        return id_map

    def connect(self):
        # These are the "Tableau 20" colors as RGB.
        tableau20: list[tuple[float, float, float]] = [
            (31, 119, 180),
            (174, 199, 232),
            (255, 127, 14),
            (255, 187, 120),
            (44, 160, 44),
            (152, 223, 138),
            (214, 39, 40),
            (255, 152, 150),
            (148, 103, 189),
            (197, 176, 213),
            (140, 86, 75),
            (196, 156, 148),
            (227, 119, 194),
            (247, 182, 210),
            (127, 127, 127),
            (199, 199, 199),
            (188, 189, 34),
            (219, 219, 141),
            (23, 190, 207),
            (158, 218, 229),
        ]

        # Scale the RGB values to the [0, 1] range, which is the format
        # matplotlib accepts.
        for i in range(len(tableau20)):
            r, g, b = tableau20[i]
            tableau20[i] = (r / 255.0, g / 255.0, b / 255.0)

        sediment_map = {
            None: np.nan,
            "loose sand": 1,
            "medium sand": 2,
            "dense sand": 3,
            "very soft clay": 4,
            "soft clay": 5,
            "firm clay": 6,
            "stiff clay": 7,
            "hard glacial till": 8,
            "cemented": 9,
            "soft rock coral": 10,
            "hard rock": 11,
            "gravel cobble": 12,
        }

        sediment_map_r = {v: k for k, v in sediment_map.items()}

        deployment_sediment = (
            self.data.deployment_strata["sediment"].sel(layer="layer 1").values
        )
        cable_corridor_sediment = (
            self.data.cable_corridor_strata["sediment"]
            .sel(layer="layer 1")
            .values
        )

        new_deployment_sediment = np.vectorize(sediment_map.get)(
            deployment_sediment
        )

        new_cable_corridor_sediment = np.vectorize(sediment_map.get)(
            cable_corridor_sediment
        )

        # Find out how many codes were used to get the levels and legend
        deployment_codes = np.unique(new_deployment_sediment)
        deployment_codes = deployment_codes[~np.isnan(deployment_codes)]
        cable_corridor_codes = np.unique(new_cable_corridor_sediment)
        cable_corridor_codes = cable_corridor_codes[
            ~np.isnan(cable_corridor_codes)
        ]

        all_codes = set(deployment_codes) | set(cable_corridor_codes)
        all_codes = list(all_codes)
        all_codes.sort()

        levels = [0] + all_codes
        legend_names = [sediment_map_r[x] for x in all_codes]
        legend_names = cast(list[str], legend_names)

        color_index = [int(x - 1) for x in all_codes]
        colors = [tableau20[x] for x in color_index]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, aspect="equal")

        x = self.data.deployment_strata.coords["x"]
        y = self.data.deployment_strata.coords["y"]

        cs = plt.contourf(
            x, y, new_deployment_sediment.T, colors=colors, levels=levels
        )

        plt.contour(
            x,
            y,
            new_deployment_sediment.T,
            levels=levels,
            colors=("0.8",),
            linewidths=(1.2,),
        )

        x = self.data.cable_corridor_strata.coords["x"]
        y = self.data.cable_corridor_strata.coords["y"]

        plt.contourf(
            x, y, new_cable_corridor_sediment.T, colors=colors, levels=levels
        )

        plt.contour(
            x,
            y,
            new_cable_corridor_sediment.T,
            levels=levels,
            colors=("0.8",),
            linewidths=(1.2,),
        )

        proxy = [Rectangle((0, 0), 1, 1, fc=cs.get_facecolor()[0])]
        lgd = ax.legend(
            proxy, legend_names, bbox_to_anchor=(1.04, 1), loc="upper left"
        )

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")

        plt.title("Combined Sediment (First Layer)")

        # Auto adjust canvas for legend
        # https://stackoverflow.com/a/45846024
        plt.gcf().canvas.draw()
        invFigure = plt.gcf().transFigure.inverted()

        lgd_pos = lgd.get_window_extent()
        lgd_coord = invFigure.transform(lgd_pos)
        lgd_xmax = lgd_coord[1, 0]

        ax_pos = plt.gca().get_window_extent()
        ax_coord = invFigure.transform(ax_pos)
        ax_xmax = ax_coord[1, 0]

        shift = ax_xmax / lgd_xmax
        plt.gcf().tight_layout(rect=(0, 0, shift, 1))

        self.fig_handle = plt.gcf()
